#!/usr/bin/env python3
"""
This is the crawler for the CompCars dataset.

It stores tuples containing the CompCars annotations for each sample.

Tuple is of the form:

(make, model, releasedyear, type, path/to/image)

- makeid: starts at 0 in tuple. Starts at 1 in dataset. See classes["make"] for human readable names
- modelid: starts at 0 in tuple. Starts at 1 in dataset. See classes["model"] for human readable names
- type: starts at 0 in tuple. Starts at 1 in dataset. See classes["type"] for human readable names
- releasedyear: No change for now...

"""
import os
import re
import glob
from typing import Dict
from scipy.io import loadmat


class CoLabelCompCarsCrawler:
    """Data crawler for CompCars Data dataset (NOT for sv-data)
    """

    makeidx = 0
    modelidx = 1
    yearidx = 2
    typeidx = 3
    pathidx = 4

    def __init__(self, data_folder="CompCars", train_folder="image", **kwargs):
        self.metadata = {}
        self.classes = {}
        self.data_folder = data_folder
        self.train_folder = os.path.join(self.data_folder, train_folder)
        self.train_file = kwargs.get("trainfile")
        self.test_file = kwargs.get("testfile")

        self.logger = kwargs.get("logger")

        self.__verify(self.data_folder)
        self.__verify(self.train_folder)

        self.crawl()

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))

    def crawl(self,):
        # self.classes = {}   #self.__getclasses(self.train_folder) # This is a map from human-readable class to numbers (i.e. the ids in the dataset...)

        # for each label we have, we label the self.matadata["train'}["classes"]["label"] = num-entities...

        # Now extract the hum,an-reasable labels...using the mat files
        human_readable = os.path.join(self.data_folder, "misc")
        makemodelmat = os.path.join(human_readable, "make_model_name.mat")
        typemat = os.path.join(human_readable, "car_type.mat")

        # Set up type humanr readable
        self.classes["type"] = {}
        self.classes["make"] = {}
        self.classes["model"] = {}
        tmat = loadmat(typemat)
        for idx in range(tmat["types"].shape[1]):  # --> 12
            self.classes["type"][idx] = tmat["types"][0, idx][0]
        mmmat = loadmat(makemodelmat)
        for idx in range(mmmat["model_names"].shape[0]):  # --> 12
            if mmmat["model_names"][idx, 0].shape[0] > 0:
                self.classes["model"][idx] = mmmat["model_names"][idx, 0][0]
            else:
                self.classes["model"][idx] = "NoName"

        for idx in range(mmmat["make_names"].shape[0]):  # --> 12
            if mmmat["make_names"][idx, 0].shape[0] > 0:
                self.classes["make"][idx] = mmmat["make_names"][idx, 0][0]
            else:
                self.classes["make"][idx] = "NoName"
        del mmmat
        del tmat

        # Get the Model-IDs to Car Type dictionary from the attributes file
        # attributes file --> (model_id maximum_speed displacement door_number seat_number type)
        model_id_type_dict = {}
        with open(
            os.path.join(os.path.join(self.data_folder, "misc"), "attributes.txt"), "r"
        ) as attr_file:
            attr_file.readline()
            for line in attr_file:
                dat = line.strip().split(" ")
                model_id_type_dict[dat[0]] = dat[-1]

        split_folder = os.path.join(self.data_folder, "train_test_split")
        class_folder = os.path.join(split_folder, "classification")

        (
            self.metadata["train"],
            self.metadata["test"],
            self.metadata["val"],
            self.metadata["full"],
        ) = ({}, {}, {}, {})
        self.metadata["train"]["crawl"], self.metadata["train"]["imgs"] = self.__crawl(
            class_folder, self.train_file, model_id_type_dict
        )
        self.metadata["test"]["crawl"], self.metadata["test"]["imgs"] = self.__crawl(
            class_folder, self.test_file, model_id_type_dict
        )
        self.metadata["val"]["crawl"], self.metadata["val"]["imgs"] = [], 0
        # full is usually handled in generator
        self.metadata["full"]["crawl"], self.metadata["full"]["imgs"] = [], 0

        # Here we need to perform some data cleaning
        # Namely, the make, model, and year ids are the raw make model
        # We need to convert them a bit, in that we need only the make model and year that exist, and have them be consistent across train and test
        self.existingmakes = {
            original: remapped
            for remapped, original in enumerate(
                set([item[self.makeidx] for item in self.metadata["train"]["crawl"]])
            )
        }
        self.existingmodels = {
            original: remapped
            for remapped, original in enumerate(
                set([item[self.modelidx] for item in self.metadata["train"]["crawl"]])
            )
        }
        self.existingyears = {
            original: remapped
            for remapped, original in enumerate(
                set([item[self.yearidx] for item in self.metadata["train"]["crawl"]])
            )
        }

        self.metadata["train"]["crawl"] = [
            [
                self.existingmakes[sampletuple[self.makeidx]],
                self.existingmodels[sampletuple[self.modelidx]],
                self.existingyears[sampletuple[self.yearidx]],
                sampletuple[self.typeidx],
                sampletuple[self.pathidx],
            ]
            for sampletuple in self.metadata["train"]["crawl"]
        ]
        self.metadata["test"]["crawl"] = [
            [
                self.existingmakes[sampletuple[self.makeidx]],
                self.existingmodels[sampletuple[self.modelidx]],
                self.existingyears[sampletuple[self.yearidx]],
                sampletuple[self.typeidx],
                sampletuple[self.pathidx],
            ]
            for sampletuple in self.metadata["test"]["crawl"]
        ]

        # Now is when we have the year information
        self.classes["year"] = self.existingyears

        # set up the necessary class information...
        for meta in ["train", "test", "val", "full"]:
            self.metadata[meta]["classes"] = {}
            self.metadata[meta]["classes"]["type"] = len(self.classes["type"])
            self.metadata[meta]["classes"]["model"] = len(self.existingmodels)
            self.metadata[meta]["classes"]["make"] = len(self.existingmakes)
            self.metadata[meta]["classes"]["year"] = len(self.classes["year"])

    def __crawl(self, folder, file, model_type_dict):

        data_tuple = (
            []
        )  # (makeid, modelid,releasedyear, type, str(path-to-imagename))   <-- EVERYTHING ELSE IS INT
        basepath = os.path.join(self.data_folder, "image")
        with open(os.path.join(folder, file), "r") as data_list:
            for line in data_list:
                parsed_line = line.strip().split("/")[:3]
                parsed_line.append(model_type_dict[parsed_line[1]])
                parsed_line = [
                    self.convert(item) - 1 for item in parsed_line
                ]  # WE RESET LABELS TO 0!!!!!!!!!!!! NOTE NOTE NOTE NOTE
                parsed_line[2] = parsed_line[2] + 1
                parsed_line.append(os.path.join(basepath, line.strip()))
                data_tuple.append(tuple(parsed_line))
        return data_tuple, len(data_tuple)
        # Now we calculate the number for each class

    def convert(self, val):
        if val == "unknown":  # For year...
            return -1
        return int(val)



from typing import Dict, List


class LabelMetadata:
    """LabelMetadata contains metadata about labels in the dataset.

    Attributes:
        labels (List[str]): A list of label names.
        totalLabels (int): The total number of labels in the LabelMetadata
        metadata (Dict[str,Dict[str,Union[str,int]]]): A dictionary of label-names to their metadata. Currently, stores the number of classes in metadata[labelname]["classes"]
    """

    labels: List[str]
    totalLabels: int
    metadata: Dict[str, Dict[str, int]]

    def __init__(self, label_dict: Dict[str, Dict[str, int]] = {}):
        """Initializes the LabelMetadata and populates it. If given an empty dict as input, the LabelMetadata will be empty

        Args:
            label_dict (Dict[str,Dict[str,Union[str,int]]], optional): A dictionary of label-names to their metadata, stored as label_dict[labelname]["classes"]. Defaults to {}.
        """
        self.labels = [*label_dict]
        self.totalLabels = len(label_dict)

        self.metadata = {
            labelname: {"classes": label_dict[labelname]["classes"],}
            for labelname in self.labels
        }

    def getLabels(self) -> List[str]:
        """Get the list of labels stored in this LabelMetadata

        Returns:
            List[str]: List of label names
        """
        return self.labels

    def getLabel(self, idx) -> str:
        """Get the label stored at this index

        Returns:
            str: Label names
        """
        return self.labels[idx]

    def getLabelDimensions(self, label: str = None) -> int:
        """Get the number of classes for a label.

        Args:
            label (str, optional): The label to retrieve number of classes for. If none, gets the first label's classes. Defaults to None.

        Returns:
            int: The number of classes
        """

        if label is None:
            return self.metadata[self.labels[0]]["classes"]
        return self.metadata[label]["classes"]

    def getTotalLabels(self) -> int:
        """Gets the total number of labels stored in this LabelMetadata

        Returns:
            int: Total number of labels
        """
        return self.totalLabels

    def addLabel(self, label: str):
        if label in self.metadata:
            raise KeyError("Label %s already in class" % label)
        self.metadata[label] = {}



from torch import nn


class InputAttention(nn.Module):
    """Input attention module for global/local attention from GLAMOR; see arXiv paper
    at https://arxiv.org/pdf/2002.02256.pdf
    """

    def __init__(self, planes):
        super(InputAttention, self).__init__()
        self.ia_conv1 = nn.Conv2d(planes, planes, kernel_size=3, padding=1, bias=False)
        self.ia_relu1 = nn.LeakyReLU()
        self.ia_conv2 = nn.Conv2d(planes, planes, kernel_size=3, padding=1, bias=False)
        self.ia_sigmoid = nn.Sigmoid()

    def forward(self, x):
        x = self.ia_conv1(x)
        x = self.ia_relu1(x)
        x = self.ia_conv2(x)
        x = self.ia_sigmoid(x)
        return x



import importlib
import torch

from ednaml.models.ModelAbstract import ModelAbstract


class BaseOptimizer:
    """ Base Optimizer Builder

    """

    name: str = "optimizer1"

    def __init__(
        self,
        name,
        optimizer,
        base_lr,
        lr_bias,
        gpus,
        weight_decay,
        weight_bias,
        opt_kwargs,
    ):
        """ Initializes the optimizer builder.

        Args:
        base_lr (float): Base learning rate for optimizer
        lr_bias (float): Multiplicative factor for bias parameters
        gpus (int): Number of GPUs for lr scaling
        weight_decay (float): Weight decay for decoupled weight decay optimizers like AdamW
        weight_bias (float): Multiplicative factor for bias parameters in weight decay optimizers

        Methods:
        build:  builds an optimizer given optimizer name and torch model

        """
        self.name = name
        self.optimizer = optimizer
        self.base_lr = base_lr
        self.gpus = gpus
        self.weight_decay = weight_decay
        self.lr_bias = lr_bias
        self.weight_bias = weight_bias
        self.kwargs = opt_kwargs

    def build(self, model: ModelAbstract) -> torch.optim.Optimizer:
        """ Builds an optimizer.

        Args:
        model (torch.nn.Module): A model
        name (str): name of torch.optim object to build
        kwargs (dict): any parameters that need to be passed into the optimizer

        Returns:
        torch.optim object

        """
        params = []
        for key, value in model.named_parameters():
            if value.requires_grad:
                if "bias" in key:
                    learning_rate = self.base_lr * self.lr_bias
                    weight_decay = self.weight_decay * self.weight_bias
                else:
                    learning_rate = self.base_lr * self.gpus
                    weight_decay = self.weight_decay
                params += [
                    {
                        "params": [value],
                        "lr": learning_rate,
                        "weight_decay": weight_decay,
                    }
                ]
        optimizer = importlib.import_module(self.optimizer, package="torch.optim")
        optimizer = optimizer(params, **self.kwargs)
        return optimizer


from .ClassificationOptimizer import ClassificationOptimizer
from .StandardLossOptimizer import StandardLossOptimizer

# for re-id compatibility



# BaseConfig class to manage the input configurations. TODO

from re import S
from typing import Dict, List
import yaml
from ednaml.config.ConfigDefaults import ConfigDefaults

from ednaml.config.ExecutionConfig import ExecutionConfig
from ednaml.config.LoggingConfig import LoggingConfig
from ednaml.config.LossConfig import LossConfig
from ednaml.config.OptimizerConfig import OptimizerConfig
from ednaml.config.SaveConfig import SaveConfig
from ednaml.config.SchedulerConfig import SchedulerConfig
from ednaml.config.TransformationConfig import TransformationConfig
from ednaml.config.ModelConfig import ModelConfig


class EdnaMLConfig:
    EXECUTION: ExecutionConfig
    SAVE: SaveConfig
    TRANSFORMATION: TransformationConfig
    MODEL: ModelConfig
    LOSS: List[LossConfig]
    OPTIMIZER: List[OptimizerConfig]  # one optimizer for each set of model params
    SCHEDULER: List[SchedulerConfig]  # one scheduler for each optimizer
    LOSS_OPTIMIZER: List[OptimizerConfig]  # one optimizer for each loss with params
    LOSS_SCHEDULER: List[SchedulerConfig]  # one scheduler for each loss_optimizer
    LOGGING: LoggingConfig

    def __init__(self, config_path: str, defaults: ConfigDefaults = ConfigDefaults()):
        ydict = {}
        with open(config_path, "r") as cfile:
            ydict = yaml.safe_load(cfile.read().strip())

        self.EXECUTION = ExecutionConfig(ydict.get("EXECUTION", {}), defaults)
        self.SAVE = SaveConfig(ydict.get("SAVE", {}), defaults)
        self.TRANSFORMATION = TransformationConfig(
            ydict.get("TRANSFORMATION", {}), defaults
        )
        self.MODEL = ModelConfig(
            ydict.get("MODEL", {}), defaults
        )  # No default MODEL itself, though it will be instantiated here? deal with this TODO
        self.LOSS = [
            LossConfig(loss_item, defaults) for loss_item in ydict.get("LOSS", [])
        ]  # No default LOSS itself

        # Default optimizer is Adam
        self.OPTIMIZER = [
            OptimizerConfig(optimizer_item, defaults)
            for optimizer_item in ydict.get("OPTIMIZER", [{}])
        ]
        self.SCHEDULER = [
            SchedulerConfig(scheduler_item, defaults)
            for scheduler_item in ydict.get("SCHEDULER", [{}])
        ]

        self.LOSS_OPTIMIZER = [
            OptimizerConfig(optimizer_item, defaults)
            for optimizer_item in ydict.get("LOSS_OPTIMIZER", [{}])
        ]
        self.LOSS_SCHEDULER = [
            SchedulerConfig(scheduler_item, defaults)
            for scheduler_item in ydict.get("LOSS_SCHEDULER", [{}])
        ]

        self.LOGGING = LoggingConfig(ydict.get("LOGGING", {}), defaults)

        # Things without defaults that MUST be provided: model ✅, train_dataloader, loss ✅, trainer TODO


"""
Notes for LOSS_OPTIMIZER and LOSS_SCHEDULER
        Now we import the loss-optimizers and loss-scheduler
        Here, we need some additional code to make things work.
        
        Situation 1. LossBuilders, plus fully enumerated loss-optimizers. Simply load them into the config
        Situation 2. LossBuilders with no learnable params, and no loss-optimizers.
        Situation 3. LossBuilders with learnable parameters, and no loss-optimizers.
        Situation 4. LossBuilders, some with learnable parameters, and only corresponding loss-optimizers
        Situation 5. LossBuilders, some with learnable parameters, and less loss-optimizers
        Situation 6. LossBuilders, some with learnable parameters, and more loss-optimizers

        To cover all cases elegantly, we will simply provide:
        Situation 1. All loss-optimizers
        Situation 2. No way to check params here. Provide a default loss-optimizer
        Situation 3. No way to check params here. Provide a default loss-optimizer
        Situation 4. Provide enumerated loss-optimizers
        Situation 5. Provide enumerated loss-optimizers
        Situation 6. Provide enumerated loss-optimizers


        Then, EdnaML will:
        - construct a temporary dictionary of loss-optimizer configs, ordered by their name
        - For each loss-builder, check if there is a loss-optimizer with that name in the dict
        - If there is, then construct optimizer with those parameters
        - If there is not, then use the first optimizer's parameters to construct missing ones
        - this way, for each situation:
            Situation 1. No problems
            Situation 2. Default parameters used, but the optimizer builder will return None
            Situation 3. Default parameters used.
            Situation 4. No problems. The LossBuilders without params will return None
            Situation 5. Missing optimizers will be replaced with the first optimizer
            Situation 6. Extra optimizers will do nothing

        For Loss-schedulers, either there are enumerated schedulers, or a single default
        in EdnaML, we will match schedulers to the corresponding optimizers
        For optimizer without scheduler, create a scheduler using the first one
"""



import tqdm, json
from sklearn.metrics import f1_score
import shutil
import os
import torch
import numpy as np
import ednaml.loss.builders
from typing import List
from ednaml.crawlers import Crawler
from ednaml.trainer import BaseTrainer
from ednaml.utils.LabelMetadata import LabelMetadata


class ClassificationTrainer(BaseTrainer):
    def __init__(
        self,
        model: torch.nn.Module,
        loss_fn: List[ednaml.loss.builders.LossBuilder],
        optimizer: torch.optim.Optimizer,
        loss_optimizer: List[torch.optim.Optimizer],
        scheduler: torch.optim.lr_scheduler._LRScheduler,
        loss_scheduler: List[torch.optim.lr_scheduler._LRScheduler],
        train_loader,
        test_loader,
        epochs: int,
        skipeval,
        logger,
        crawler: Crawler,
        config,
        labels: LabelMetadata,
        **kwargs
    ):

        super().__init__(
            model,
            loss_fn,
            optimizer,
            loss_optimizer,
            scheduler,
            loss_scheduler,
            train_loader,
            test_loader,
            epochs,
            skipeval,
            logger,
            crawler,
            config,
            labels,
            **kwargs
        )

        self.softaccuracy = []

    # The train function for the CoLabel model is inherited

    def epoch_step(self, epoch):
        """Trains for an epoch
        """
        for batch in self.train_loader:
            # Set up scheduled LR
            if self.global_batch == 0:
                lrs = self.scheduler.get_last_lr()
                lrs = sum(lrs) / float(len(lrs))
                self.logger.info(
                    "Starting epoch {0} with {1} steps and learning rate {2:2.5E}".format(
                        epoch,
                        len(self.train_loader) - (len(self.train_loader) % 10),
                        lrs,
                    )
                )

            # Step through the batch (including loss.backward())
            self.step(batch)
            self.global_batch += 1

            if (self.global_batch + 1) % self.step_verbose == 0:
                loss_avg = 0.0
                for lossname in self.loss_fn:
                    loss_avg += (
                        sum(self.losses[lossname][-self.step_verbose :])
                        / self.step_verbose
                    )
                loss_avg /= self.num_losses
                soft_avg = sum(self.softaccuracy[-100:]) / float(
                    len(self.softaccuracy[-100:])
                )
                self.logger.info(
                    "Epoch{0}.{1}\tTotal Avg Loss: {2:.3f} Softmax: {3:.3f}".format(
                        self.global_epoch, self.global_batch, loss_avg, soft_avg
                    )
                )

        self.global_batch = 0

        # Step the lr schedule to update the learning rate
        self.scheduler.step()
        for lossname in self.loss_fn:
            if self.loss_scheduler[lossname] is not None:
                self.loss_scheduler[lossname].step()

        self.logger.info(
            "{0} Completed epoch {1} {2}".format("*" * 10, self.global_epoch, "*" * 10)
        )

        if self.global_epoch % self.test_frequency == 0:
            self.logger.info("Evaluating model at test-frequency")
            self.evaluate()
        if self.global_epoch % self.save_frequency == 0:
            self.logger.info("Saving model at save-frequency")
            self.save()
        self.global_epoch += 1

    # Steps through a batch of data
    def step(self, batch):
        # Switch the model to training mode
        self.model.train()
        self.optimizer.zero_grad()
        for lossname in self.loss_fn:
            if self.loss_optimizer[lossname] is not None:
                self.loss_optimizer[lossname].zero_grad()

        batch_kwargs = {}
        (
            img,
            batch_kwargs["labels"],
        ) = batch  # This is the tensor response from collate_fn
        img, batch_kwargs["labels"] = img.cuda(), batch_kwargs["labels"].cuda()
        # logits, features, labels
        batch_kwargs["logits"], batch_kwargs["features"], _ = self.model(img)
        batch_kwargs["epoch"] = self.global_epoch  # For CompactContrastiveLoss

        # TODO fix this with info about how many output are in the model...from the config file!!!!!
        loss = {loss_name: None for loss_name in self.loss_fn}
        for lossname in self.loss_fn:
            loss[lossname] = self.loss_fn[lossname](**batch_kwargs)
        # if self.fp16 and self.apex is not None:
        #    with self.apex.amp.scale_loss(loss, self.optimizer) as scaled_loss:
        #        scaled_loss.backward()
        # else:
        #    loss.backward()
        lossbackward = sum(loss.values())
        lossbackward.backward()

        self.optimizer.step()
        for lossname in self.loss_fn:
            if (
                self.loss_optimizer[lossname] is not None
            ):  # In case loss object doesn;t have any parameters, this will be None. See optimizers.StandardLossOptimizer
                self.loss_optimizer[lossname].step()

        for idx, lossname in enumerate(self.loss_fn):
            self.losses[lossname].append(loss[lossname].cpu().item())

        if batch_kwargs["logits"] is not None:
            softmax_accuracy = (
                (batch_kwargs["logits"].max(1)[1] == batch_kwargs["labels"])
                .float()
                .mean()
            )
            self.softaccuracy.append(softmax_accuracy.cpu().item())
        else:
            self.softaccuracy.append(0)

    def evaluate_impl(self):
        self.model.eval()
        features, logits, labels = [], [], []
        with torch.no_grad():
            for batch in tqdm.tqdm(
                self.test_loader, total=len(self.test_loader), leave=False
            ):
                data, label = batch
                data = data.cuda()
                logit, feature, _ = self.model(data)
                feature = feature.detach().cpu()
                logit = logit.detach().cpu()
                features.append(feature)
                logits.append(logit)
                labels.append(label)

        features, logits, labels = (
            torch.cat(features, dim=0),
            torch.cat(logits, dim=0),
            torch.cat(labels, dim=0),
        )
        # Now we compute the loss...
        self.logger.info("Obtained features, validation in progress")
        # for evaluation...

        logit_labels = torch.argmax(logits, dim=1)
        accuracy = (logit_labels == labels).sum().float() / float(labels.size(0))
        micro_fscore = np.mean(f1_score(labels, logit_labels, average="micro"))
        weighted_fscore = np.mean(f1_score(labels, logit_labels, average="weighted"))
        self.logger.info("Accuracy: {:.3%}".format(accuracy))
        self.logger.info("Micro F-score: {:.3f}".format(micro_fscore))
        self.logger.info("Weighted F-score: {:.3f}".format(weighted_fscore))
        return logit_labels, labels, features

    def saveMetadata(self,):
        self.logger.info("Saving model metadata")
        jMetadata = json.dumps(self.metadata)
        metafile = "metadata.json"
        localmetafile = os.path.join(self.save_directory, metafile)
        if self.save_backup:
            backupmetafile = os.path.join(self.backup_directory, metafile)
        if not os.path.exists(localmetafile):
            with open(localmetafile, "w") as localmetaobj:
                localmetaobj.write(jMetadata)
        self.logger.info("Backing up metadata")
        if self.save_backup:
            shutil.copy2(localmetafile, backupmetafile)
        self.logger.info("Finished metadata backup")



from torch import nn
from ednaml.utils.blocks import InputAttention


class ResnetInput(nn.Module):
    def __init__(self, ia_attention: bool = False) -> None:
        super().__init__()

        self.inplanes = 64
        self.ia = ia_attention
        self.conv1 = nn.Conv2d(
            3, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False
        )
        # if norm_layer == "gn":
        #    self.bn1 = nn.GroupNorm2d
        self.bn1 = nn.BatchNorm2d(self.inplanes)
        self.relu1 = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.ia_attention = nn.Identity()
        if self.ia:
            self.ia_attention = InputAttention(self.inplanes)  # 64, set above

    def forward(self, x):
        x = self.conv1(x)
        if self.ia:
            x = self.ia_attention(x) * x
        x = self.bn1(x)
        x = self.relu1(x)
        x = self.maxpool(x)
        return x



import torch
from bisect import bisect_right


class FineGrainedSteppedLR(torch.optim.lr_scheduler._LRScheduler):
    """FineGrainedSteppedLR allows definition of additively or multiplicative changes to LR (either increasing or decreasing).

    Args:
        -- optimizer: torch.optim object
        -- lr_ops: list of tuples. Each tuple is a 3-tuple of (milestone, op, val).
            -- milestone: int for epoch when an op is triggered
            -- op: is a string, either '+' or '*' for addition or multiplication
            -- val: is a float that will act as an operand with the 'op' operator with the current learning rate.
        -- last_epoch: Previous epoch if resume state or launching new optimizer state during training
        -- relative: `bool`, whether milestones are relative to last_epoch or absolute epoch milestones.

            Example: 
                base_lr = 1. 
                lr_ops = [(1, '+', 0.1), (3, '*', 2), (4, '+', -0.1)]
                milestones = [1,3,4]

                lr_schedule:
                    Epoch 0 -- 1.0      previous_lr = base_lr
                    Epoch 1 -- 1.1      previous_lr = previous_lr '+' 0.1
                    Epoch 2 -- 1.1      no change
                    Epoch 3 -- 2.2      previous_lr = previous_lr '*' 1.0
                    Epoch 4 -- 2.1      previous_lr = previous_lr '+' -0.1
    """

    def __init__(self, optimizer, last_epoch=-1, **kwargs):
        lr_ops = kwargs.get("lr_ops")
        milestones = [item[0] for item in lr_ops]
        if milestones != sorted(milestones):
            raise ValueError("`milestones` in `lr_ops` should be a sorted `list`.")

        relative = kwargs.get("relative", False)
        self.build_lr = [(-1, self.lr_mult, 1)]

        for _idx, _item in enumerate(lr_ops):
            _milestone, _op, _val = _item
            if relative:
                _milestone += last_epoch + 1
            if _op == "+":
                _op = self.lr_plus
            elif _op == "*":
                _op = self.lr_mult
            else:
                raise ValueError(
                    "Unknown operation token {0}. Expected one of '+', '*'.".format(
                        _item[0]
                    )
                )
            self.build_lr.append((_milestone, _op, _val))

        self.milestones = milestones
        self.lr_ops = lr_ops
        self.lr_idx = last_epoch
        # if relative milestones, then keep initial last_epoch, and use last_epoch-initial_last_epoch as comparison. Else use last_epoch as comparison.
        # use bisect_right to perform binary search to get the correct value using this comparison variable
        super(FineGrainedSteppedLR, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        if self.last_epoch == -1:
            self.lr_idx = 0
            lr_idx = 0
        else:
            lr_idx = bisect_right(self.milestones, self.last_epoch)
        if lr_idx == self.lr_idx:
            return self.base_lrs
        else:
            self.lr_idx = lr_idx
            self.base_lrs = [
                self.build_lr[lr_idx][1](base_lr, self.build_lr[lr_idx][2])
                for base_lr in self.base_lrs
            ]
            return self.base_lrs

    def lr_mult(self, val, operand):
        return val * operand

    def lr_plus(self, val, operand):
        return val + operand



import torch
from ednaml.loss import Loss


class SoftmaxLogitsLoss(Loss):
    """Standard softmax loss

    Calculates the softmax loss.

    Methods: 
        __call__: Returns loss given logits and labels.

    """

    def forward(self, logits, labels):
        """
    Args:
        logits: prediction matrix (before softmax) with shape (batch_size, softmax_dimensions)
        labels: ground truth labels with shape (batch_size)
    """
        return torch.nn.functional.cross_entropy(
            logits[labels >= 0], labels[labels >= 0]
        )



import torch
from torch import nn


class SpatialAttention(nn.Module):
    """Spatial Attention module that forms part of CBAM attention.
    See the ECCV paper at https://openaccess.thecvf.com/content_ECCV_2018/papers/Sanghyun_Woo_Convolutional_Block_Attention_ECCV_2018_paper.pdf
    """

    def __init__(self, kernel_size=7):
        super(SpatialAttention, self).__init__()

        assert kernel_size in (3, 7), "kernel size must be 3 or 7"
        padding = 3 if kernel_size == 7 else 1

        self.conv1 = nn.Conv2d(2, 1, kernel_size, padding=padding, bias=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = torch.mean(x, dim=1, keepdim=True)
        max_out, _ = torch.max(x, dim=1, keepdim=True)
        x = torch.cat([avg_out, max_out], dim=1)
        x = self.conv1(x)
        return self.sigmoid(x)



from collections import defaultdict
import os
import re
import glob
import xml.etree.ElementTree as ET


class VeRiDataCrawler:
    """ Data crawler for the VeRi-776 dataset

    The VeRiDataCrawler crawls the VeRi-776 folder to populate training, query, and gallery sets with images and their respective Pid and Cid.

    Args:
      data_folder (str): Name of the VeRi folder. Default: "VeRi"
      train_folder (str): Folder inside data_folder with training images. Default: "image_train"
      test_folder (str): Folder inside data_folder with testing/gallery images. Default: "image_test"
      query_folder (str): Folder inside data_folder with query images. Default: "image_query"
    
    Kwargs:
      logger: Instance of Logging object

    Attributes:
      metadata (dict): Contains image paths, Pid, and Cid of training, testing, and query sets
        train (dict): contains image paths, Pid, and Cid
          crawl (list): List of tuples. Each tuple is a 5-tuple of (image_path, PID, CID, color, type)
          pid (int): Number of Pid in training set
          cid (int): Number of Cid in training set
          imgs (int): Number of images in training set
        test (dict): contains image paths, Pid, and Cid
          crawl (list): List of tuples. Each tuple is a 5-tuple of (image_path, PID, CID, color, type)
          pid (int): Number of Pid in testing set
          cid (int): Number of Cid in testing set
          imgs (int): Number of images in testing set
        query (dict): contains image paths, Pid, and Cid
          crawl (list): List of tuples. Each tuple is a 5-tuple of (image_path, PID, CID, color, type)
          pid (int): Number of Pid in query set
          cid (int): Number of Cid in query set
          imgs (int): Number of images in query set

    Methods:
      crawl(): Populate self.metadata
      __verify(folder): Check if folder exists

    """

    def __init__(
        self,
        data_folder="VeRi",
        train_folder="image_train",
        test_folder="image_test",
        query_folder="image_query",
        **kwargs
    ):
        self.metadata = {}

        self.data_folder = data_folder
        self.train_folder = os.path.join(self.data_folder, train_folder)
        self.test_folder = os.path.join(self.data_folder, test_folder)
        self.query_folder = os.path.join(self.data_folder, query_folder)
        self.tracks_file = os.path.join(self.data_folder, "test_track.txt")
        self.testlabel = os.path.join(self.data_folder, "test_label.xml")
        self.trainlabel = os.path.join(self.data_folder, "train_label.xml")

        self.logger = kwargs.get("logger")

        self.__verify(self.data_folder)
        self.__verify(self.train_folder)
        self.__verify(self.test_folder)
        self.__verify(self.query_folder)

        self.crawl()

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))

    def crawl(self,):
        self.colordict = defaultdict(lambda: -1)
        self.typedict = defaultdict(lambda: -1)
        with open(self.trainlabel, "r") as rfile:
            traintree = ET.fromstring(rfile.read())
        # traintree = ET.parse(self.trainlabel, parser=ET.XMLParser(encoding='gb2312'))

        for item in traintree[0]:
            vid = item.get("vehicleID")
            colorid = item.get("colorID")
            typeid = item.get("typeID")
            self.colordict[int(vid)] = int(colorid) - 1
            self.typedict[int(vid)] = int(typeid) - 1
        # traintree = ET.parse(self.testlabel, parser=ET.XMLParser(encoding='gb2312'))
        with open(self.testlabel, "r") as rfile:
            traintree = ET.fromstring(rfile.read())
        # trainroot = traintree.find("Items")
        for item in traintree[0]:
            vid = item.get("vehicleID")
            colorid = item.get("colorID")
            typeid = item.get("typeID")
            self.colordict[int(vid)] = int(colorid) - 1
            self.typedict[int(vid)] = int(typeid) - 1
        del traintree

        self.classes = {}
        self.classes["color"] = 10
        self.classes["type"] = 9

        (
            self.metadata["train"],
            self.metadata["test"],
            self.metadata["query"],
            self.metadata["track"],
        ) = ({}, {}, {}, {})
        (
            self.metadata["train"]["crawl"],
            self.metadata["train"]["pid"],
            self.metadata["train"]["cid"],
            self.metadata["train"]["imgs"],
        ) = self.__crawl(self.train_folder, reset_labels=True)
        (
            self.metadata["test"]["crawl"],
            self.metadata["test"]["pid"],
            self.metadata["test"]["cid"],
            self.metadata["test"]["imgs"],
        ) = self.__crawl(self.test_folder)
        (
            self.metadata["query"]["crawl"],
            self.metadata["query"]["pid"],
            self.metadata["query"]["cid"],
            self.metadata["query"]["imgs"],
        ) = self.__crawl(self.query_folder)

        (
            self.metadata["track"]["crawl"],
            self.metadata["track"]["pid"],
            self.metadata["track"]["cid"],
            self.metadata["track"]["imgs"],
            self.metadata["track"]["dict"],
            self.metadata["track"]["info"],
        ) = self.__crawltracks(self.test_folder)

        # Extras for colabel...
        self.metadata["val"], self.metadata["full"] = {}, {}
        self.metadata["val"]["crawl"], self.metadata["full"]["crawl"] = [], []
        for meta in ["train", "test", "val", "full"]:
            self.metadata[meta]["imgs"] = len(self.metadata[meta]["crawl"])
            self.metadata[meta]["classes"] = {}
            self.metadata[meta]["classes"]["color"] = 10
            self.metadata[meta]["classes"]["type"] = 9
            self.metadata[meta]["classes"]["pid"] = self.metadata["train"]["pid"]
            self.metadata[meta]["classes"]["cid"] = self.metadata["train"]["cid"]

        self.logger.info(
            "Train\tPID: {:6d}\tCID: {:6d}\tIMGS: {:8d}".format(
                self.metadata["train"]["pid"],
                self.metadata["train"]["cid"],
                self.metadata["train"]["imgs"],
            )
        )
        self.logger.info(
            "Test \tPID: {:6d}\tCID: {:6d}\tIMGS: {:8d}".format(
                self.metadata["test"]["pid"],
                self.metadata["test"]["cid"],
                self.metadata["test"]["imgs"],
            )
        )
        self.logger.info(
            "Query\tPID: {:6d}\tCID: {:6d}\tIMGS: {:8d}".format(
                self.metadata["query"]["pid"],
                self.metadata["query"]["cid"],
                self.metadata["query"]["imgs"],
            )
        )
        self.logger.info(
            "Tracks\tPID: {:6d}\tCID: {:6d}\Tracks: {:8d}".format(
                self.metadata["track"]["pid"],
                self.metadata["track"]["cid"],
                self.metadata["track"]["imgs"],
            )
        )

    def __crawl(self, folder, reset_labels=False):
        imgs = glob.glob(os.path.join(folder, "*.jpg"))
        _re = re.compile(r"([\d]+)_[a-z]([\d]+)")
        pid_labeler = 0
        pid_tracker, cid_tracker = {}, {}
        crawler = []
        pid_counter, cid_counter, img_counter = 0, 0, 0
        for img in imgs:
            pid, cid = map(int, _re.search(img).groups())  # _re.search lol
            if pid < 0:
                continue  # ignore junk
            if cid < 0:
                continue  # ignore junk
            if pid not in pid_tracker:
                pid_tracker[pid] = pid_labeler if reset_labels else pid
                pid_labeler += 1
            if cid not in cid_tracker:
                cid_tracker[cid] = cid - 1
            crawler.append(
                (
                    img,
                    pid_tracker[pid],
                    cid - 1,
                    self.colordict[pid],
                    self.typedict[pid],
                )
            )  # cid start at 1 in data
        return crawler, len(pid_tracker), len(cid_tracker), len(crawler)

    def __crawltracks(self, folder, reset_labels=False):

        _re = re.compile(r"([\d]+)_[a-z]([\d]+)")
        pid_labeler = 0
        pid_tracker, cid_tracker = {}, {}
        crawler = []
        pid_counter, cid_counter, img_counter = 0, 0, 0
        track_dict, track_info = {}, {}
        track_idx = 0
        with open(self.tracks_file, "r") as tracks_file_reader:
            for line in tracks_file_reader:
                # each line is a track...
                track_list = line.strip().split(" ")
                track_index = track_list[0]
                track_images = track_list[1:]
                track_images = [os.path.join(folder, item) for item in track_images]
                pid, cid = map(
                    int, _re.search(track_images[0]).groups()
                )  # _re.search lol
                if pid < 0:
                    continue  # ignore junk
                if cid < 0:
                    continue  # ignore junk
                if pid not in pid_tracker:
                    pid_tracker[pid] = pid_labeler if reset_labels else pid
                    pid_labeler += 1
                crawler.append(
                    (
                        track_images,
                        pid_tracker[pid],
                        cid - 1,
                        self.colordict[pid],
                        self.typedict[pid],
                    )
                )  # cid start at 1 in data
                # if len(crawler) == 1650:
                for img in track_images:
                    track_dict[img] = track_idx
                track_info[track_idx] = {"pid": pid_tracker[pid], "cid": cid - 1}
                track_idx += 1
        return (
            crawler,
            len(pid_tracker),
            len(cid_tracker),
            len(crawler),
            track_dict,
            track_info,
        )



import tqdm, json
from sklearn.metrics import f1_score
import shutil
import os
import torch
import numpy as np
import ednaml.loss.builders
from typing import List
from ednaml.models import MultiBranchResnet
from ednaml.crawlers import Crawler
from ednaml.trainer import BaseTrainer
from ednaml.utils.LabelMetadata import LabelMetadata


class MultiBranchTrainer(BaseTrainer):
    model: MultiBranchResnet

    def __init__(
        self,
        model: MultiBranchResnet,
        loss_fn: List[ednaml.loss.builders.LossBuilder],
        optimizer: torch.optim.Optimizer,
        loss_optimizer: List[torch.optim.Optimizer],
        scheduler: torch.optim.lr_scheduler._LRScheduler,
        loss_scheduler: torch.optim.lr_scheduler._LRScheduler,
        train_loader,
        test_loader,
        epochs: int,
        skipeval,
        logger,
        crawler: Crawler,
        config,
        labels: LabelMetadata,
        **kwargs
    ):

        super().__init__(
            model,
            loss_fn,
            optimizer,
            loss_optimizer,
            scheduler,
            loss_scheduler,
            train_loader,
            test_loader,
            epochs,
            skipeval,
            logger,
            crawler,
            config,
            labels,
            **kwargs
        )

        self.softaccuracy = []

        # mapping label names and class names to their index for faster retrieval.
        # TODO some way to integrate classificationclass in DATAREADER to
        # labelnames in MODEL, so that there is less redundancy...
        self.model_labelorder = {
            item: idx for idx, item in enumerate(self.model.model_labelorder)
        }
        self.model_nameorder = {
            item: idx for idx, item in enumerate(self.model.model_nameorder)
        }
        self.data_labelorder = {
            item: idx for idx, item in enumerate(self.labelMetadata.labels)
        }
        self.model_name_label_map = {
            item[0]: item[1]
            for item in zip(self.model.model_nameorder, self.model.model_labelorder)
        }

    # The train function for the CoLabel model is inherited

    def epoch_step(self, epoch):
        """Trains for an epoch
        """
        for batch in self.train_loader:
            # Set up scheduled LR
            if self.global_batch == 0:
                lrs = self.scheduler.get_last_lr()
                lrs = sum(lrs) / float(len(lrs))
                self.logger.info(
                    "Starting epoch {0} with {1} steps and learning rate {2:2.5E}".format(
                        epoch,
                        len(self.train_loader) - (len(self.train_loader) % 10),
                        lrs,
                    )
                )

            # Step through the batch (including loss.backward())
            self.step(batch)
            self.global_batch += 1

            if (self.global_batch + 1) % self.step_verbose == 0:
                loss_avg = 0.0
                for lossname in self.losses:
                    loss_avg += (
                        sum(self.losses[lossname][-self.step_verbose :])
                        / self.step_verbose
                    )
                loss_avg /= self.num_losses
                soft_avg = sum(self.softaccuracy[-100:]) / float(
                    len(self.softaccuracy[-100:])
                )
                self.logger.info(
                    "Epoch{0}.{1}\tTotal Avg Loss: {2:.3f} Softmax: {3:.3f}".format(
                        self.global_epoch, self.global_batch, loss_avg, soft_avg
                    )
                )

        self.global_batch = 0

        # Step the lr schedule to update the learning rate
        self.scheduler.step()
        for lossname in self.loss_fn:
            if self.loss_scheduler[lossname] is not None:
                self.loss_scheduler[lossname].step()

        self.logger.info(
            "{0} Completed epoch {1} {2}".format("*" * 10, self.global_epoch, "*" * 10)
        )

        if self.global_epoch % self.test_frequency == 0:
            self.logger.info("Evaluating model at test-frequency")
            self.evaluate()
        if self.global_epoch % self.save_frequency == 0:
            self.logger.info("Saving model at save-frequency")
            self.save()
        self.global_epoch += 1

    # Steps through a batch of data
    def step(self, batch):
        # Switch the model to training mode
        self.model.train()
        self.optimizer.zero_grad()
        for lossname in self.loss_fn:
            if self.loss_optimizer[lossname] is not None:
                self.loss_optimizer[lossname].zero_grad()

        batch_kwargs = {}
        (
            img,
            batch_kwargs["labels"],
        ) = batch  # This is the tensor response from collate_fn
        img, batch_kwargs["labels"] = (
            img.cuda(),
            batch_kwargs["labels"].cuda(),
        )  # labels are in order of labelnames
        # logits, features, labels
        batch_kwargs["logits"], batch_kwargs["features"], _ = self.model(
            img
        )  # logits are in order of output_classnames --> model.output_classnames
        batch_kwargs["epoch"] = self.global_epoch  # For CompactContrastiveLoss

        loss = {loss_name: None for loss_name in self.loss_fn}
        for lossname in loss:  # this loss targets a specific output
            akwargs = {}

            akwargs["logits"] = batch_kwargs["logits"][
                self.model_labelorder[self.model_name_label_map[lossname]]
            ]  # this looks up the lossname in the outputclass names
            akwargs["labels"] = batch_kwargs["labels"][
                :, self.data_labelorder[self.model_name_label_map[lossname]]
            ]  # ^ditto
            akwargs["epoch"] = batch_kwargs["epoch"]
            loss[lossname] = self.loss_fn[lossname](**akwargs)

        lossbackward = sum(loss.values())
        lossbackward.backward()

        # for idx in range(self.num_losses):
        #    loss[idx].backward()

        self.optimizer.step()
        for lossname in self.loss_fn:
            if (
                self.loss_optimizer[lossname] is not None
            ):  # In case loss object doesn;t have any parameters, this will be None. See optimizers.StandardLossOptimizer
                self.loss_optimizer[lossname].step()

        for idx, lossname in enumerate(self.loss_fn):
            self.losses[lossname].append(loss[lossname].cpu().item())

        # if batch_kwargs["logits"] is not None:
        # softmax_accuracy = (batch_kwargs["logits"].max(1)[1] == batch_kwargs["labels"]).float().mean()
        # self.softaccuracy.append(softmax_accuracy.cpu().item())
        # else:
        self.softaccuracy.append(0)  # TODO fix this

    def evaluate_impl(self):
        self.model.eval()
        features, logits, labels = (
            [[] for _ in range(self.model.feature_count)],
            [[] for _ in range(self.model.output_count)],
            [],
        )
        with torch.no_grad():
            for batch in tqdm.tqdm(
                self.test_loader, total=len(self.test_loader), leave=False
            ):
                data, label = batch
                data = data.cuda()
                logit, feature, _ = self.model(data)
                for idx in range(self.model.feature_count):
                    features[idx].append(feature[idx].detach().cpu())
                for idx in range(self.model.output_count):
                    logits[idx].append(logit[idx].detach().cpu())
                labels.append(label)

        # features, logits, labels = torch.cat(features, dim=0), [torch.cat(logit, dim=0) for logit in logits], torch.cat(labels, dim=0)
        features = [torch.cat(feature, dim=0) for feature in features]
        logits = [torch.cat(logit, dim=0) for logit in logits]
        labels = torch.cat(labels, dim=0)
        # Now we compute the loss...
        self.logger.info("Obtained features, validation in progress")
        # for evaluation...

        logit_labels = [torch.argmax(logit, dim=1) for logit in logits]
        accuracy = [[] for _ in range(self.model.output_count)]
        micro_fscore = [[] for _ in range(self.model.output_count)]
        weighted_fscore = [[] for _ in range(self.model.output_count)]

        """
        akwargs["logits"] = batch_kwargs["logits"][self.model_labelorder[self.model_name_label_map[lossname]]] # this looks up the lossname in the outputclass names
        akwargs["labels"] = batch_kwargs["labels"][:, self.data_labelorder[self.model_name_label_map[lossname]]] # ^ditto
        """
        for idx, lossname in enumerate(self.loss_fn):
            accuracy[idx] = (
                logit_labels[self.model_labelorder[self.model_name_label_map[lossname]]]
                == labels[:, self.data_labelorder[self.model_name_label_map[lossname]]]
            ).sum().float() / float(labels.size(0))
            micro_fscore[idx] = np.mean(
                f1_score(
                    labels[
                        :, self.data_labelorder[self.model_name_label_map[lossname]]
                    ],
                    logit_labels[
                        self.model_labelorder[self.model_name_label_map[lossname]]
                    ],
                    average="micro",
                )
            )
            weighted_fscore[idx] = np.mean(
                f1_score(
                    labels[
                        :, self.data_labelorder[self.model_name_label_map[lossname]]
                    ],
                    logit_labels[
                        self.model_labelorder[self.model_name_label_map[lossname]]
                    ],
                    average="weighted",
                )
            )

        self.logger.info(
            "Metrics\t" + "\t".join(["%s" % lossname for lossname in self.loss_fn])
        )
        self.logger.info(
            "Accuracy\t"
            + "\t".join(
                [
                    "%s: %0.3f"
                    % (self.model.model_nameorder[idx], accuracy[idx].item())
                    for idx in range(self.model.output_count)
                ]
            )
        )
        self.logger.info(
            "M F-Score\t"
            + "\t".join(
                [
                    "%s: %0.3f"
                    % (self.model.model_nameorder[idx], micro_fscore[idx].item())
                    for idx in range(self.model.output_count)
                ]
            )
        )
        self.logger.info(
            "W F-Score\t"
            + "\t".join(
                [
                    "%s: %0.3f"
                    % (self.model.model_nameorder[idx], weighted_fscore[idx].item())
                    for idx in range(self.model.output_count)
                ]
            )
        )

        return logit_labels, labels, features

    def saveMetadata(self,):
        self.logger.info("Saving model metadata")
        jMetadata = json.dumps(self.metadata)
        metafile = "metadata.json"
        localmetafile = os.path.join(self.save_directory, metafile)
        if self.save_backup:
            backupmetafile = os.path.join(self.backup_directory, metafile)
        if not os.path.exists(localmetafile):
            with open(localmetafile, "w") as localmetaobj:
                localmetaobj.write(jMetadata)
        self.logger.info("Backing up metadata")
        if self.save_backup:
            shutil.copy2(localmetafile, backupmetafile)
        self.logger.info("Finished metadata backup")



from torch import nn
import torch

from utils.LabelMetadata import LabelMetadata


class ModelAbstract(nn.Module):
    model_name = "ModelAbstract"
    model_arch = None
    number_outputs = 1
    output_classnames = ["out1"]
    output_dimensions = [512]
    secondary_outputs = []

    model_base: str
    weights: str
    normalization: str
    metadata: LabelMetadata

    def __init__(
        self,
        base=None,
        weights=None,
        metadata: LabelMetadata = None,
        normalization=None,
        **kwargs
    ):
        super().__init__()
        self.metadata: LabelMetadata = metadata
        self.model_base = base
        self.weights = weights
        self.normalization = normalization

        self.model_attributes_setup(**kwargs)
        self.model_setup(**kwargs)

    def model_attributes_setup(self, **kwargs):
        raise NotImplementedError()

    def model_setup(self, **kwargs):
        raise NotImplementedError()

    def weights_init_kaiming(self, m):
        classname = m.__class__.__name__
        if classname.find("Linear") != -1:
            nn.init.kaiming_normal_(m.weight, a=0, mode="fan_out")
            nn.init.constant_(m.bias, 0.0)
        elif classname.find("Conv") != -1:
            nn.init.kaiming_normal_(m.weight, a=0, mode="fan_in")
            if m.bias is not None:
                nn.init.constant_(m.bias, 0.0)
        elif classname.find("BatchNorm") != -1:
            if m.affine:
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)
        elif classname.find("GroupNorm") != -1:
            if m.affine:
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)
        elif classname.find("LayerNorm") != -1:
            # if m.affine:
            #    nn.init.constant_(m.weight, 1.0)
            #    nn.init.constant_(m.bias, 0.0)
            pass
        elif classname.find("InstanceNorm") != -1:
            if m.affine:
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)

    def weights_init_softmax(self, m):
        """ Initialize linear weights to standard normal. Mean 0. Standard Deviation 0.001 """
        classname = m.__class__.__name__
        if classname.find("Linear") != -1:
            nn.init.normal_(m.weight, std=0.001)
            if m.bias:
                nn.init.constant_(m.bias, 0.0)

    def partial_load(self, weights_path):
        params = torch.load(weights_path)
        for _key in params:
            if (
                _key not in self.state_dict().keys()
                or params[_key].shape != self.state_dict()[_key].shape
            ):
                continue
            self.state_dict()[_key].copy_(params[_key])

    def forward(self, x, **kwargs):
        feature_logits, features, secondary_outputs = self.forward_impl(x, **kwargs)

        return feature_logits, features, secondary_outputs

    def foward_impl(self, x, **kwargs):

        raise NotImplementedError()

    def getModelName(self):
        return self.model_name

    def getModelBase(self):
        return self.model_base

    def getModelArch(self):
        return self.model_arch



from ednaml.loss.builders import LossBuilder

from ednaml.loss import SoftmaxLogitsLoss
from ednaml.loss import SoftmaxLabelSmooth


class ClassificationLossBuilder(LossBuilder):
    LOSS_PARAMS = {}
    LOSS_PARAMS["SoftmaxLogitsLoss"] = {}
    LOSS_PARAMS["SoftmaxLogitsLoss"]["fn"] = SoftmaxLogitsLoss
    LOSS_PARAMS["SoftmaxLogitsLoss"]["args"] = ["logits", "labels"]
    LOSS_PARAMS["SoftmaxLabelSmooth"] = {}
    LOSS_PARAMS["SoftmaxLabelSmooth"]["fn"] = SoftmaxLabelSmooth
    LOSS_PARAMS["SoftmaxLabelSmooth"]["args"] = ["logits", "labels"]



from typing import List
from torch import nn
import warnings
from ednaml.utils.LabelMetadata import LabelMetadata


class LossBuilder(nn.Module):
    LOSS_PARAMS = {}
    loss: nn.ModuleList
    loss_labelname: str
    loss_label: str
    loss_classes_metadata: LabelMetadata
    loss_lambda: List[int]

    def __init__(self, loss_functions, loss_lambda, loss_kwargs, **kwargs):
        super().__init__()
        self.loss = nn.ModuleList([])

        fn_len = len(loss_functions)
        lambda_len = len(loss_lambda)
        kwargs_len = len(loss_kwargs)

        self.loss_label = kwargs.get("label", "color")

        self.loss_labelname = kwargs.get("name", None)
        if self.loss_labelname is None:
            warnings.warn(
                "No name given for this loss, using label as name %s" % self.loss_label
            )
            self.loss_labelname = self.loss_label

        self.loss_classes_metadata: LabelMetadata = kwargs.get("metadata")

        # Set up the logger
        self.logger = kwargs.get("logger")
        # Sanity check
        if fn_len != lambda_len:
            raise ValueError(
                "Loss function list length is %i. Expected %i length loss_lambdas, got %i"
                % (fn_len, fn_len, lambda_len)
            )
        if fn_len != kwargs_len:
            raise ValueError(
                "Loss function list length is %i. Expected %i length loss_kwargs, got %i"
                % (fn_len, fn_len, kwargs_len)
            )
        # Add the loss functions with the correct features. Lambda is applied later
        for idx, loss_fn_name in enumerate(loss_functions):
            self.loss.append(
                self.LOSS_PARAMS[loss_fn_name]["fn"](
                    lossname=self.loss_label,
                    metadata=self.loss_classes_metadata,
                    **loss_kwargs[idx]
                )
            )
            self.logger.info(
                "Added {loss} with lambda = {lamb} and loss arguments {largs}".format(
                    loss=loss_fn_name,
                    lamb=loss_lambda[idx],
                    largs=str(loss_kwargs[idx]),
                )
            )

        lambda_sum = sum(loss_lambda)
        loss_lambda = [float(item) / float(lambda_sum) for item in loss_lambda]
        self.loss_lambda = loss_lambda
        self.loss_fn = loss_functions

    def forward(self, **kwargs):
        """Call operator of the loss builder.

        This returns the sum of each individual loss provided in the initialization, multiplied by their respective loss_lambdas. 
        TODO update this + CarZam base model forward to deal with logits as well if necessary

        Args (kwargs only):
            labels: Torch tensor of shape (batch_size, 1). The class labels.
            features: Torch tensor of shape (batch_size, embedding_dimensions). The feature embeddings generated by the ReID model.
        """
        loss = 0.0
        for idx, loss_fn in enumerate(self.loss):
            # loss += self.loss_lambda[idx] * fn(kwargs.get(self.LOSS_PARAMS[self.loss_fn[idx]]['args'][0]), kwargs.get(self.LOSS_PARAMS[self.loss_fn[idx]]['args'][1]), kwargs.get(self.LOSS_PARAMS[self.loss_fn[idx]]['args'][2]))
            loss += self.loss_lambda[idx] * loss_fn(
                *[
                    kwargs.get(arg_name)
                    for arg_name in self.LOSS_PARAMS[self.loss_fn[idx]]["args"]
                ]
            )

        return loss


from .ClassificationLossBuilder import ClassificationLossBuilder



from typing import List, Dict, Tuple
import os
import logging


def dynamic_import(cfg, module_name: str, import_name: str, default: str = None):
    """ Perform a dynamic import 

    Args:
        cfg (kaptan Kaptan object): A kaptan configuration object
        module_name (str): The directory where the module-to-import resides. For example to import a crawler, module_name should be 'crawlers'
        import_name (str): The key in the kaptan configuration that provides the import name
        default (str): If import_name does not exist, `default` will be used instead

    Returns:
        Imported module attribute

    Raises:
        ModuleNotFoundError (implicit): If provided module is not found
        AttributeError (implicit): If provided import_name or default does not exist
    """

    import_name = cfg.get(import_name, default)
    imported_module = __import__(
        "%s." % module_name + import_name, fromlist=[import_name]
    )
    return getattr(imported_module, import_name)


def extend_mean_arguments(
    params_to_fix: List[str] = [0.5, 0.5], channels=3
) -> Tuple[List[float]]:
    return tuple([[item] * channels for item in params_to_fix])


def generate_logger(MODEL_SAVE_FOLDER, LOGGER_SAVE_NAME):
    logger = logging.getLogger(MODEL_SAVE_FOLDER)
    if logger.hasHandlers():
        return logger

    logger.setLevel(logging.DEBUG)
    logger_save_path = os.path.join(MODEL_SAVE_FOLDER, LOGGER_SAVE_NAME)

    fh = logging.FileHandler(logger_save_path)
    fh.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s-%(msecs)d %(message)s", datefmt="%H:%M:%S"
    )
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    cs = logging.StreamHandler()
    cs.setLevel(logging.DEBUG)
    cs.setFormatter(
        logging.Formatter("%(asctime)s-%(msecs)d %(message)s", datefmt="%H:%M:%S")
    )
    logger.addHandler(cs)
    return logger


def generate_save_names(cfg):
    MODEL_SAVE_NAME = "%s-v%i" % (
        cfg.get("SAVE.MODEL_CORE_NAME"),
        cfg.get("SAVE.MODEL_VERSION"),
    )
    MODEL_SAVE_FOLDER = "%s-v%i-%s-%s" % (
        cfg.get("SAVE.MODEL_CORE_NAME"),
        cfg.get("SAVE.MODEL_VERSION"),
        cfg.get("SAVE.MODEL_BACKBONE"),
        cfg.get("SAVE.MODEL_QUALIFIER"),
    )
    LOGGER_SAVE_NAME = "%s-v%i-%s-%s-logger.log" % (
        cfg.get("SAVE.MODEL_CORE_NAME"),
        cfg.get("SAVE.MODEL_VERSION"),
        cfg.get("SAVE.MODEL_BACKBONE"),
        cfg.get("SAVE.MODEL_QUALIFIER"),
    )
    if cfg.get("SAVE.DRIVE_BACKUP"):
        CHECKPOINT_DIRECTORY = (
            cfg.get("SAVE.CHECKPOINT_DIRECTORY", "./drive/My Drive/Vehicles/Models/")
            + MODEL_SAVE_FOLDER
        )
    else:
        CHECKPOINT_DIRECTORY = ""
    return MODEL_SAVE_NAME, MODEL_SAVE_FOLDER, LOGGER_SAVE_NAME, CHECKPOINT_DIRECTORY


def fix_generator_arguments(
    cfg: Dict, params_to_fix: List[str] = []
) -> Tuple[List[float]]:
    if len(params_to_fix) > 0:
        return_params = [None] * len(params_to_fix)
        for idx, param in enumerate(params_to_fix):
            if type(cfg.get(param)) is int or type(cfg.get(param)) is float:
                return_params[idx] = [cfg.get(param)] * cfg.get(
                    "TRANSFORMATION.CHANNELS"
                )
        return tuple(return_params)
    else:
        if (
            type(cfg.get("TRANSFORMATION.NORMALIZATION_MEAN")) is int
            or type(cfg.get("TRANSFORMATION.NORMALIZATION_MEAN")) is float
        ):
            NORMALIZATION_MEAN = [
                cfg.get("TRANSFORMATION.NORMALIZATION_MEAN")
            ] * cfg.get("TRANSFORMATION.CHANNELS")
        if (
            type(cfg.get("TRANSFORMATION.NORMALIZATION_STD")) is int
            or type(cfg.get("TRANSFORMATION.NORMALIZATION_STD")) is float
        ):
            NORMALIZATION_STD = [cfg.get("TRANSFORMATION.NORMALIZATION_STD")] * cfg.get(
                "TRANSFORMATION.CHANNELS"
            )
        if (
            type(cfg.get("TRANSFORMATION.RANDOM_ERASE_VALUE")) is int
            or type(cfg.get("TRANSFORMATION.RANDOM_ERASE_VALUE")) is float
        ):
            RANDOM_ERASE_VALUE = [
                cfg.get("TRANSFORMATION.RANDOM_ERASE_VALUE")
            ] * cfg.get("TRANSFORMATION.CHANNELS")
        return NORMALIZATION_MEAN, NORMALIZATION_STD, RANDOM_ERASE_VALUE


model_weights = {
    "resnet18": [
        "https://download.pytorch.org/models/resnet18-5c106cde.pth",
        "resnet18-5c106cde.pth",
    ],
    "resnet34": [
        "https://download.pytorch.org/models/resnet34-333f7ec4.pth",
        "resnet34-333f7ec4.pth",
    ],
    "resnet50": [
        "https://download.pytorch.org/models/resnet50-19c8e357.pth",
        "resnet50-19c8e357.pth",
    ],
    "resnet101": [
        "https://download.pytorch.org/models/resnet101-5d3b4d8f.pth",
        "resnet101-5d3b4d8f.pth",
    ],
    "resnet152": [
        "https://download.pytorch.org/models/resnet152-b121ed2d.pth",
        "resnet152-b121ed2d.pth",
    ],
    "resnext50_32x4d": [
        "https://download.pytorch.org/models/resnext50_32x4d-7cdf4587.pth",
        "resnext50_32x4d-7cdf4587.pth",
    ],
    "resnext101_32x8d": [
        "https://download.pytorch.org/models/resnext101_32x8d-8ba56ff5.pth",
        "resnext101_32x8d-8ba56ff5.pth",
    ],
    "wide_resnet50_2": [
        "https://download.pytorch.org/models/wide_resnet50_2-95faca4d.pth",
        "wide_resnet50_2-95faca4d.pth",
    ],
    "wide_resnet101_2": [
        "https://download.pytorch.org/models/wide_resnet101_2-32ee1156.pth",
        "wide_resnet50_2-95faca4d.pth",
    ],
    "resnet18_cbam": [
        "https://download.pytorch.org/models/resnet18-5c106cde.pth",
        "resnet18-5c106cde_cbam.pth",
    ],
    "resnet34_cbam": [
        "https://download.pytorch.org/models/resnet34-333f7ec4.pth",
        "resnet34-333f7ec4_cbam.pth",
    ],
    "resnet50_cbam": [
        "https://download.pytorch.org/models/resnet50-19c8e357.pth",
        "resnet50-19c8e357_cbam.pth",
    ],
    "resnet101_cbam": [
        "https://download.pytorch.org/models/resnet101-5d3b4d8f.pth",
        "resnet101-5d3b4d8f_cbam.pth",
    ],
    "resnet152_cbam": [
        "https://download.pytorch.org/models/resnet152-b121ed2d.pth",
        "resnet152-b121ed2d_cbam.pth",
    ],
    "shufflenetv2_small": [
        "https://github.com/asuprem/shufflenet-models/raw/master/ShuffleNetV2%2B.Small.pth.tar",
        "shufflenetv2-small.pth",
    ],
}



from typing import Dict, List

from ednaml.config.ConfigDefaults import ConfigDefaults


class LossConfig:
    LOSSES: List[str]
    KWARGS: List[Dict[str, str]]
    LAMBDAS: List[int]
    LABEL: str
    NAME: str

    def __init__(self, loss_dict, defaults: ConfigDefaults):
        self.LOSSES = loss_dict.get("LOSSES", defaults.LOSSES)
        self.KWARGS = loss_dict.get("KWARGS", defaults.LOSS_KWARGS)
        self.LAMBDAS = loss_dict.get("LAMBDAS", defaults.LAMBDAS)
        self.LABEL = loss_dict.get("LABEL", defaults.LOSS_LABEL)
        self.NAME = loss_dict.get("NAME", defaults.LOSS_NAME)



from itertools import combinations
import numpy as np
import torch
import torch.nn.functional as F
from ednaml.loss import Loss


class ContrastiveLoss(Loss):
    """Standard contrastive loss

    Calculates the contrastive loss of a mini-batch.

    Args (kwargs only):
        margin (float, 0.3): Margin constraint to use in triplet loss. If not provided,
        mine (str): Mining method. Default 'hard'. Supports ['hard', 'all']. 

    Methods: 
        __call__: Returns loss given features and labels.
    """

    def __init__(self, **kwargs):
        super(ContrastiveLoss, self).__init__()
        self.margin = kwargs.get("margin", 0.3)

    def forward(self, features, labels):
        positive_pairs, negative_pairs = self.get_pairs(features, labels)
        if features.is_cuda:
            positive_pairs = positive_pairs.cuda()
            negative_pairs = negative_pairs.cuda()
        positive_loss = (
            (features[positive_pairs[:, 0]] - features[positive_pairs[:, 1]])
            .pow(2)
            .sum(1)
        )
        negative_loss = F.relu(
            self.margin
            - (features[negative_pairs[:, 0]] - features[negative_pairs[:, 1]])
            .pow(2)
            .sum(1)
            .sqrt()
        ).pow(2)
        loss = torch.cat([positive_loss, negative_loss], dim=0)
        return loss.mean()

    def get_pairs(self, features, labels):
        distance_matrix = self.pdist(features)

        labels = labels.cpu().data.numpy()
        all_pairs = np.array(list(combinations(range(len(labels)), 2)))
        all_pairs = torch.LongTensor(all_pairs)
        positive_pairs = all_pairs[
            (labels[all_pairs[:, 0]] == labels[all_pairs[:, 1]]).nonzero()
        ]
        negative_pairs = all_pairs[
            (labels[all_pairs[:, 0]] != labels[all_pairs[:, 1]]).nonzero()
        ]

        negative_distances = distance_matrix[negative_pairs[:, 0], negative_pairs[:, 1]]
        negative_distances = negative_distances.cpu().data.numpy()
        top_negatives = np.argpartition(negative_distances, len(positive_pairs))[
            : len(positive_pairs)
        ]
        top_negative_pairs = negative_pairs[torch.LongTensor(top_negatives)]

        return positive_pairs, top_negative_pairs

    def pdist(self, vectors):
        distance_matrix = (
            -2 * vectors.mm(torch.t(vectors))
            + vectors.pow(2).sum(dim=1).view(1, -1)
            + vectors.pow(2).sum(dim=1).view(-1, 1)
        )
        return distance_matrix



from ednaml.config.ConfigDefaults import ConfigDefaults


class SaveConfig:
    MODEL_VERSION: int
    MODEL_CORE_NAME: str
    MODEL_BACKBONE: str
    MODEL_QUALIFIER: str
    DRIVE_BACKUP: bool
    SAVE_FREQUENCY: int
    CHECKPOINT_DIRECTORY: str

    def __init__(self, save_dict, defaults: ConfigDefaults):
        self.MODEL_VERSION = save_dict.get("MODEL_VERSION", defaults.MODEL_VERSION)
        self.MODEL_CORE_NAME = save_dict.get(
            "MODEL_CORE_NAME", defaults.MODEL_CORE_NAME
        )
        self.MODEL_BACKBONE = save_dict.get("MODEL_BACKBONE", defaults.MODEL_BACKBONE)
        self.MODEL_QUALIFIER = save_dict.get(
            "MODEL_QUALIFIER", defaults.MODEL_QUALIFIER
        )
        self.DRIVE_BACKUP = save_dict.get("DRIVE_BACKUP", defaults.DRIVE_BACKUP)
        self.SAVE_FREQUENCY = save_dict.get("SAVE_FREQUENCY", defaults.SAVE_FREQUENCY)
        self.CHECKPOINT_DIRECTORY = save_dict.get(
            "CHECKPOINT_DIRECTORY", defaults.CHECKPOINT_DIRECTORY
        )



from collections import defaultdict
import os
import random

# Tuple:
# (imgpath, pid, cid, color, model)


class VehicleIDDataCrawler:
    def __init__(
        self,
        data_folder="VehicleID",
        train_folder="image",
        test_folder="",
        query_folder="",
        **kwargs
    ):
        self.metadata = {}

        self.data_folder = data_folder
        self.image_folder = os.path.join(self.data_folder, train_folder)
        self.attribute_folder = os.path.join(
            self.data_folder, kwargs.get("attribute_folder", "attribute")
        )

        self.train_list = "train_list.txt"
        self.query_list = kwargs.get("test_list", "test_list_2400") + ".txt"

        list_folder = os.path.join(self.data_folder, "train_test_split")
        # The train list is in VehicleID/train_test_split/train_list.txt
        # The gallery/query list is in VehicleID/train_test_split/test_list_13164.txt
        self.train_list = os.path.join(list_folder, self.train_list)
        self.query_list = os.path.join(list_folder, self.query_list)

        self.logger = kwargs.get("logger")

        self.__verify(self.data_folder)
        self.__verify(self.image_folder)

        self.crawl()

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))

    def crawl(self,):

        self.colordict = defaultdict(lambda: -1)
        with open(os.path.join(self.attribute_folder, "color_attr.txt"), "r") as cfile:
            for line in cfile:
                line = line.strip().split(" ")
                self.colordict[int(line[0])] = int(
                    line[1]
                )  # colordict[vehicleid]=colorid
        self.modeldict = defaultdict(lambda: -1)
        with open(os.path.join(self.attribute_folder, "model_attr.txt"), "r") as cfile:
            for line in cfile:
                line = line.strip().split(" ")
                self.modeldict[int(line[0])] = int(
                    line[1]
                )  # modeldict[vehicleid]=modelid

        self.classes = {}
        self.classes["color"] = 7
        self.classes["model"] = 250

        self.metadata["train"], self.metadata["gallery"], self.metadata["query"] = (
            {},
            {},
            {},
        )
        (
            self.metadata["train"]["crawl"],
            self.metadata["train"]["pids"],
            self.metadata["train"]["cids"],
            self.metadata["train"]["imgs"],
        ) = self.__crawl(self.train_list, reset_labels=True)

        self.metadata["test"] = {}
        # self.metadata["test"]["crawl"] = [self.metadata["train"]["crawl"].pop(random.randrange(len(self.metadata["train"]["crawl"]))) for _ in range(int(len(self.metadata["train"]["crawl"])*0.1))]
        random.shuffle(self.metadata["train"]["crawl"])
        nshuffle = int(len(self.metadata["train"]["crawl"]) * 0.1)
        self.metadata["test"]["crawl"] = self.metadata["train"]["crawl"][:nshuffle]
        self.metadata["train"]["crawl"] = self.metadata["train"]["crawl"][nshuffle:]
        self.metadata["train"]["imgs"] = len(self.metadata["train"]["crawl"])
        self.metadata["test"]["imgs"] = len(self.metadata["test"]["crawl"])

        self.__querycrawl(self.query_list)

        # Extras for colabel...
        self.metadata["val"], self.metadata["full"] = {}, {}
        self.metadata["val"]["crawl"], self.metadata["full"]["crawl"] = [], []
        for meta in ["train", "gallery", "val", "full", "test"]:
            self.metadata[meta]["imgs"] = len(self.metadata[meta]["crawl"])
            self.metadata[meta]["classes"] = {}
            self.metadata[meta]["classes"]["color"] = 7
            self.metadata[meta]["classes"]["make"] = 250

        self.logger.info(
            "Train\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["train"]["pids"],
                self.metadata["train"]["cids"],
                self.metadata["train"]["imgs"],
            )
        )
        self.logger.info(
            "Gallery \tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["gallery"]["pids"],
                self.metadata["gallery"]["cids"],
                self.metadata["gallery"]["imgs"],
            )
        )
        self.logger.info(
            "Query\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["query"]["pids"],
                self.metadata["query"]["cids"],
                self.metadata["query"]["imgs"],
            )
        )

    def __crawl(self, train_file, reset_labels=False):
        crawler = []
        pids, cids = {}, []
        pid_label = 0
        with open(train_file, "r") as train_file_reader:
            for line in train_file_reader:
                ln = line.strip().split(" ")
                img_path = ln[0] + ".jpg"
                img_path = os.path.join(self.image_folder, img_path)
                pid = int(ln[1])
                cid = 0
                if pid not in pids:
                    pids[pid] = pid_label if reset_labels else pid
                    pid_label += 1
                # pids.append(pid)
                cids.append(cid)
                crawler.append(
                    (img_path, pids[pid], cid, self.colordict[pid], self.modeldict[pid])
                )
        return crawler, len(set(pids.keys())), len(set(cids)), len(crawler)

    def __querycrawl(self, query_file, reset_labels=False):
        crawler = []
        pids, cids = {}, []
        pid_label = 0
        with open(query_file, "r") as query_file_reader:
            for line in query_file_reader:
                ln = line.strip().split(" ")
                img_path = ln[0] + ".jpg"
                img_path = os.path.join(self.image_folder, img_path)
                pid = int(ln[1])
                cid = 0
                if pid not in pids:
                    pids[pid] = pid_label if reset_labels else pid
                    pid_label += 1
                # pids.append(pid)
                cids.append(cid)
                crawler.append((img_path, pids[pid], cid))

        pid_in_gallery = {}
        self.metadata["gallery"]["crawl"], self.metadata["query"]["crawl"] = [], []

        for crawled_img in crawler:
            img_path, pid, cid = crawled_img
            # check if pid already captured. If so add to query. Else add to gallery (based on paper) (variable pid_in_gallery should be pid_in)gallery
            if pid in pid_in_gallery:
                self.metadata["query"]["crawl"].append(
                    (img_path, pid, cid, self.colordict[pid], self.modeldict[pid])
                )
            else:
                pid_in_gallery[pid] = 1
                self.metadata["gallery"]["crawl"].append(
                    (img_path, pid, cid, self.colordict[pid], self.modeldict[pid])
                )

        self.metadata["gallery"]["pids"], self.metadata["gallery"]["cids"] = (
            len(pids),
            1,
        )
        self.metadata["query"]["pids"], self.metadata["query"]["cids"] = len(pids), 1

        self.metadata["gallery"]["imgs"] = len(self.metadata["gallery"]["crawl"])
        self.metadata["query"]["imgs"] = len(self.metadata["query"]["crawl"])



import os


class VRICDataCrawler:
    """ Data crawler for the VRIC (Vehicle Re-identification in Context) dataset

    The VRIC crawls the VRIC data folder to populate training, query, and gallery sets with images and their respective PIDs and CIDs. It uses the provided list of training, query, and gallery images to populate the crawler metadata.

    Args:
      data_folder (str): Name of the VeRi folder. Default: "VRIC"
      train_folder (str): Folder inside data_folder with training images. Default: "train_images"
      test_folder (str): Folder inside data_folder with testing/gallery images. Default: "gallery_images"
      query_folder (str): Folder inside data_folder with query images. Default: "probe_images"
    
    Kwargs:
      logger: Instance of Logging object

    Attributes:
      metadata (dict): Contains image paths, PIDs, and CIDs of training, testing, and query sets
        train (dict): contains image paths, PIDs, and CIDs
          crawl (list): List of tuples. Each tuple is a 3-tuple of (image_path, PID, CID)
          pids (int): Number of PIDs in training set
          cids (int): Number of CIDs in training set
          imgs (int): Number of images in training set
        test (dict): contains image paths, PIDs, and CIDs
          crawl (list): List of tuples. Each tuple is a 3-tuple of (image_path, PID, CID)
          pids (int): Number of PIDs in testing set
          cids (int): Number of CIDs in testing set
          imgs (int): Number of images in testing set
        query (dict): contains image paths, PIDs, and CIDs
          crawl (list): List of tuples. Each tuple is a 3-tuple of (image_path, PID, CID)
          pids (int): Number of PIDs in query set
          cids (int): Number of CIDs in query set
          imgs (int): Number of images in query set

    Methods:
      crawl(): Populate self.metadata
      __verify(folder): Check if folder exists

    """

    def __init__(
        self,
        data_folder="VRIC",
        train_folder="train_images",
        test_folder="gallery_images",
        query_folder="probe_images",
        **kwargs
    ):
        self.metadata = {}

        self.data_folder = data_folder
        self.train_folder = os.path.join(self.data_folder, train_folder)
        self.test_folder = os.path.join(self.data_folder, test_folder)
        self.query_folder = os.path.join(self.data_folder, query_folder)

        self.logger = kwargs.get("logger")

        self.__verify(self.data_folder)
        self.__verify(self.train_folder)
        self.__verify(self.test_folder)
        self.__verify(self.query_folder)

        self.crawl()

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))

    def crawl(self,):
        # We will build the details using the text files
        self.metadata["train"], self.metadata["test"], self.metadata["query"] = (
            {},
            {},
            {},
        )
        (
            self.metadata["train"]["crawl"],
            self.metadata["train"]["pids"],
            self.metadata["train"]["cids"],
            self.metadata["train"]["imgs"],
        ) = self.__crawl(
            os.path.join(self.data_folder, "vric_train.txt"),
            self.train_folder,
            reset_labels=True,
        )
        (
            self.metadata["test"]["crawl"],
            self.metadata["test"]["pids"],
            self.metadata["test"]["cids"],
            self.metadata["test"]["imgs"],
        ) = self.__crawl(
            os.path.join(self.data_folder, "vric_gallery.txt"),
            self.test_folder,
            reset_labels=True,
        )
        (
            self.metadata["query"]["crawl"],
            self.metadata["query"]["pids"],
            self.metadata["query"]["cids"],
            self.metadata["query"]["imgs"],
        ) = self.__crawl(
            os.path.join(self.data_folder, "vric_probe.txt"),
            self.query_folder,
            reset_labels=True,
        )

        self.logger.info(
            "Train\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["train"]["pids"],
                self.metadata["train"]["cids"],
                self.metadata["train"]["imgs"],
            )
        )
        self.logger.info(
            "Test \tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["test"]["pids"],
                self.metadata["test"]["cids"],
                self.metadata["test"]["imgs"],
            )
        )
        self.logger.info(
            "Query\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(
                self.metadata["query"]["pids"],
                self.metadata["query"]["cids"],
                self.metadata["query"]["imgs"],
            )
        )

    def __crawl(self, file_to_open, folder, reset_labels=False):
        pid_labeler = 0
        pid_tracker, cid_tracker = {}, {}
        crawler = []
        pid_counter, cid_counter, img_counter = 0, 0, 0

        with open(file_to_open, "r") as file_:
            for line in file_:
                path, pid, cid = line.strip().split(" ")
                pid = int(pid)
                cid = int(cid)
                path = os.path.join(folder, path)
                if pid < 0:
                    continue  # ignore junk
                if cid < 0:
                    continue  # ignore junk
                if pid not in pid_tracker:
                    pid_tracker[pid] = pid_labeler if reset_labels else pid
                    pid_labeler += 1
                if cid not in cid_tracker:
                    cid_tracker[cid] = cid - 1
                crawler.append(
                    (path, pid_tracker[pid], cid - 1)
                )  # cids start at 1 in data
            return crawler, len(pid_tracker), len(cid_tracker), len(crawler)



import os


class CoLabelIntegratedDatasetCrawler:
    """Data crawler for CoLabel Integrated Data dataset
    """

    def __init__(
        self,
        data_folder="CoDataset",
        train_folder="image",
        test_folder="",
        validation_folder="",
        **kwargs
    ):
        self.metadata = {}
        self.classes = {}
        self.annotations = {}
        self.data_folder = data_folder
        self.train_folder = os.path.join(self.data_folder, train_folder)

        self.logger = kwargs.get("logger")

        self.__verify(self.data_folder)
        self.__verify(self.train_folder)

        self.crawl()

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))

    def crawl(self,):
        # self.classes = {}   #self.__getclasses(self.train_folder) # This is a map from human-readable class to numbers (i.e. the ids in the dataset...)

        # for each label we have, we label the self.matadata["train'}["classes"]["label"] = num-entities...

        # Now extract the hum,an-reasable labels...using the mat files
        human_readable_directory = os.path.join(self.data_folder, "readable")

        # Set up type human readable
        annotations = ["make", "color", "type"]
        for annotation in annotations:
            self.extract_readable_class_names(
                annotation, human_readable_directory, self.classes
            )  # into self.classes...
            self.extract_readable_class_names(
                annotation, human_readable_directory, self.annotations
            )  # into self.classes...
        self.extract_readable_class_names(
            "model", human_readable_directory, self.classes
        )  # into self.classes...

        split_folder = os.path.join(self.data_folder, "splits")
        class_folder = os.path.join(split_folder, "vmmr")

        self.metadata["train"], self.metadata["test"], self.metadata["val"] = {}, {}, {}
        self.metadata["train"]["crawl"], self.metadata["train"]["imgs"] = self.__crawl(
            class_folder, "train.txt"
        )
        self.metadata["test"]["crawl"], self.metadata["test"]["imgs"] = self.__crawl(
            class_folder, "test'txt"
        )
        # self.metadata["val"]["crawl"], self.metadata["val"]["classes"], self.metadata["val"]["imgs"] = self.__crawl(self.val_folder)

        self.metadata["train"]["classes"] = {}
        self.metadata["train"]["classes"]["type"] = len(self.classes["type"])
        self.metadata["train"]["classes"]["model"] = len(self.classes["model"])
        self.metadata["train"]["classes"]["make"] = len(self.classes["make"])
        self.metadata["train"]["classes"]["color"] = len(self.classes["color"])
        self.metadata["test"]["classes"] = {}
        self.metadata["test"]["classes"]["type"] = len(self.classes["type"])
        self.metadata["test"]["classes"]["model"] = len(self.classes["model"])
        self.metadata["test"]["classes"]["make"] = len(self.classes["make"])
        self.metadata["test"]["classes"]["color"] = len(self.classes["color"])

    def extract_readable_class_names(self, annotation, dir, storage):
        """Extracts the class names for make, model, etc from the respective txt files. Each line
        corresponds to a class id, and the string is the human-readable format.

        Args:
            annotation (str): The annotation to retrieve. It is in the [annotation].txt file
            dir (str): The directory where these are kept
            storage (Dict): Which dict to store these annotations
        """
        idx = 0
        storage[annotation] = {}
        with open(os.path.join(dir, annotation + ".txt"), "r") as ofile:
            for line in ofile:
                storage[annotation][idx] = line.strip()
            idx += 1

    def __crawl(self, folder, file, model_type_dict):

        data_tuple = (
            []
        )  # (makeid, modelid,typeid, colorid, str(imagename))   <-- EVERYTHING ELSE IS INT
        basepath = os.path.join(self.data_folder, "image")
        with open(os.path.join(folder, file), "r") as data_list:
            for line in data_list:
                parsed_line = line.strip().split("/")[
                    :4
                ]  # <type>/<color>/[78/1/3ac218c0c6c378.jpg] --> [typeid,colorid,makeid,modelid,path.jpg]
                parsed_line = [
                    int(item) - 1 for item in parsed_line
                ]  # WE RESET LABELS TO 0!!!!!!!!!!!! NOTE NOTE NOTE NOTE
                parsed_line.append(
                    os.path.join(basepath, "/".join(parsed_line[2:]))
                )  #  makeid,modelid,path.jpg --> makeid/modelid/path.jpg
                data_tuple.append(tuple(parsed_line))
        return data_tuple, len(data_tuple)
        # Now we calculate the number for each class



import os
import glob
from typing import Dict


class CoLabelVehicleColorCrawler:
    """Data crawler for VehicleColor dataset

    
    """

    def __init__(
        self,
        data_folder="VehicleColor",
        train_folder="train",
        test_folder="test",
        validation_folder="val",
        logger=None,
        **kwargs
    ):
        self.metadata = {}
        self.classes = {}
        self.data_folder = data_folder
        self.train_folder = os.path.join(self.data_folder, train_folder)
        self.test_folder = os.path.join(self.data_folder, test_folder)
        self.val_folder = os.path.join(self.data_folder, validation_folder)

        if logger is None:
            raise ValueError("Must pass a logger instance that is not None")
        self.logger = logger

        self.__verify(self.data_folder)
        self.__verify(self.train_folder)
        self.__verify(self.test_folder)

        self.crawl()

    def __verify(self, folder):
        if not os.path.exists(folder):
            raise IOError(
                "Folder {data_folder} does not exist".format(data_folder=folder)
            )
        else:
            self.logger.info("Found {data_folder}".format(data_folder=folder))

    def crawl(self,):

        self.classes["color"] = self.__getclasses(self.train_folder)
        # Class fixings...
        # grey, brown, gold, tan, beige, silver, black, green, orange, blue, purple, pink, white, red, yellow
        # We will merge (brown, tan, beige) --> brown
        # Merge grey, silver --> grey
        # Merge gold --> yellow

        adj = 4
        netvals = len(self.classes["color"]) - adj  # Should be 11
        adjustments = ["gold", "tan", "beige", "silver"]
        originals = [item for item in self.classes["color"] if item not in adjustments]
        for idx, original in enumerate(originals):
            # here, idx is the adjusted index...
            self.classes["color"][original] = idx

        self.classes["color"]["gold"] = self.classes["color"]["yellow"]
        self.classes["color"]["tan"] = self.classes["color"]["brown"]
        self.classes["color"]["beige"] = self.classes["color"]["brown"]
        self.classes["color"]["silver"] = self.classes["color"]["grey"]

        self.metadata["train"], self.metadata["test"], self.metadata["val"] = {}, {}, {}
        (
            self.metadata["train"]["classes"],
            self.metadata["test"]["classes"],
            self.metadata["val"]["classes"],
        ) = ({}, {}, {})
        (
            self.metadata["train"]["crawl"],
            self.metadata["train"]["classes"]["color"],
            self.metadata["train"]["imgs"],
        ) = self.__crawl(self.train_folder, adj=4)
        (
            self.metadata["test"]["crawl"],
            self.metadata["test"]["classes"]["color"],
            self.metadata["test"]["imgs"],
        ) = self.__crawl(self.test_folder, adj=4)
        (
            self.metadata["val"]["crawl"],
            self.metadata["val"]["classes"]["color"],
            self.metadata["val"]["imgs"],
        ) = self.__crawl(self.val_folder, adj=4)

        # self.logger.info("Train\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(self.metadata["train"]["pids"], self.metadata["train"]["cids"], self.metadata["train"]["imgs"]))
        # self.logger.info("Test \tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(self.metadata["test"]["pids"], self.metadata["test"]["cids"], self.metadata["test"]["imgs"]))
        # self.logger.info("Query\tPIDS: {:6d}\tCIDS: {:6d}\tIMGS: {:8d}".format(self.metadata["query"]["pids"], self.metadata["query"]["cids"], self.metadata["query"]["imgs"]))

        # result --> grey, brown, black, green, orange, blue, purple, pink, white, red, yellow

    def __getclasses(self, folder) -> Dict:
        class_dir_list = glob.glob(folder + "/*")
        # NOTE might not work for windows :/
        return {item.split("/")[-1]: idx for idx, item in enumerate(class_dir_list)}

    def __crawl(self, folder, adj=0):
        crawler = []
        for class_name in self.classes["color"]:
            class_path = os.path.join(folder, class_name)
            imgs = glob.glob(os.path.join(class_path, "*.jpg"))
            imgs = [(item, self.classes["color"][class_name]) for item in imgs]
            crawler += imgs
        return crawler, len(self.classes["color"]) - adj, len(crawler)



import torch
from torch import nn
from ednaml.loss import Loss


class MarginLoss(Loss):
    """Standard triplet loss

    Calculates the margin loss of a mini-batch.

    Args (kwargs only):
        margin (float): Margin constraint to use in triplet liming. If not provided, loss uses torch.nn.SoftMarginLoss. Else uses nn.SoftMarginLoss.
        mine (str): Mining method. Default 'hard'. Supports ['hard', 'all']. 

    Methods: 
        __call__: Returns loss given features and labels.
    """

    def __init__(self, **kwargs):

        super(MarginLoss, self).__init__()
        self.margin = kwargs.get("margin", None)
        mine = kwargs.get("mine", "hard")

        if self.margin is None:
            self.loss_fn = nn.SoftMarginLoss()
            print("Using SoftMarginLoss")
        else:
            self.loss_fn = nn.MarginRankingLoss(margin=self.margin)
            print("Using Margin Loss with margin=%f" % self.margin)
        if mine == "hard":
            self.mine = self.hard_mine
            print("Using batch hard mining")
        elif mine == "all":
            self.mine = self.average_mine
            print("Using batch all mining")
        else:
            raise NotImplementedError()

    def forward(self, features, labels):
        """
    Args:
        features: features matrix with shape (batch_size, feat_dim)
        labels: ground truth labels with shape (batch_size)
    """
        distances = self.euclidean_dist(features, features)
        distances_pos, distances_neg = self.mine(distances, labels)
        y = distances_neg.new().resize_as_(distances_neg).fill_(1)
        if self.margin is None:
            loss = self.loss_fn(distances_neg - distances_pos, y)
        else:
            loss = self.loss_fn(distances_neg, distances_pos, y)
        return loss

    def euclidean_dist(self, x, y):
        m, n = x.size(0), y.size(0)
        xx = torch.pow(x, 2).sum(1, keepdim=True).expand(m, n)
        yy = torch.pow(y, 2).sum(1, keepdim=True).expand(n, m).t()
        dist = xx + yy
        dist.addmm_(1, -2, x, y.t())
        dist = dist.clamp(min=1e-12).sqrt()  # for numerical stability
        return dist

    def hard_mine(self, distances, labels):
        N = distances.size(0)
        # shape [N, N]
        pos = labels.expand(N, N).eq(labels.expand(N, N).t())
        neg = labels.expand(N, N).ne(labels.expand(N, N).t())

        # `dist_ap` means distance(anchor, positive)
        # both `dist_ap` and `relative_p_inds` with shape [N, 1]
        dist_ap, _ = torch.max(distances[pos].contiguous().view(N, -1), 1, keepdim=True)
        # `dist_an` means distance(anchor, negative)
        # both `dist_an` and `relative_n_inds` with shape [N, 1]
        dist_an, _ = torch.min(distances[neg].contiguous().view(N, -1), 1, keepdim=True)
        # shape [N]
        dist_ap = dist_ap.squeeze(1)
        dist_an = dist_an.squeeze(1)
        return dist_ap, dist_an

    def average_mine(self, distances, labels):
        N = distances.size(0)
        # shape [N, N]
        pos = labels.expand(N, N).eq(labels.expand(N, N).t())
        neg = labels.expand(N, N).ne(labels.expand(N, N).t())

        # `dist_ap` means distance(anchor, positive)
        # both `dist_ap` and `relative_p_inds` with shape [N, 1]
        dist_ap, _ = torch.mean(
            distances[pos].contiguous().view(N, -1), 1, keepdim=True
        )
        # `dist_an` means distance(anchor, negative)
        # both `dist_an` and `relative_n_inds` with shape [N, 1]
        dist_an, _ = torch.mean(
            distances[neg].contiguous().view(N, -1), 1, keepdim=True
        )
        # shape [N]
        dist_ap = dist_ap.squeeze(1)
        dist_an = dist_an.squeeze(1)
        return dist_ap, dist_an



"""Contains code to build a multi-branch resnet, with a mix of
weight-shared and independent branches.

`multibranchresnet` creates subbranches contained all or part of resnet. 
Branching occurs at a specified resnet block, after which the remaining 
blocks occur on their own branches with no weight sharing. The branched 
model also has an option for branch fusing, to concatenate features.

  Typical usage example:

  model = multibranchresnet()
  x = torch.randn((3,100,100))
  features = model(x)
"""

import os
from torch import nn
import torch
from typing import Dict, List
from ednaml.utils.blocks import ResnetInput, ResnetBasicBlock, ResnetBottleneck


class multibranchresnet(nn.Module):
    """`multibranchresnet` creates a resnet with specified branches. 
    
    `multibranchresnet` creates subbranches contained all or part of resnet. 
    Branching occurs at a specified resnet block, after which the remaining 
    blocks occur on their own branches with no weight sharing. The branched 
    model also has an option for branch fusing, to concatenate features.

    Attributes:
        block (nn.Module): The block (BasicBlock or Bottleneck) used for the internal Resnet architecture
        attention (str): Which type of attentio, if any, is implemented in this model. One of `cbam`, `dbam`
        input_attention (bool): Whether model uses input attention at the first resnet block
        ia_attention (bool): Whether the model uses input attention at the first layer
        part_attention (bool): Whether the model uses local/part attention at the first resnet block
        secondary_attention (None | int): Whether `attention` is applied to all blocks (None) or to the specified block (int). 
        shared_block_count (int): Number of resnet blocks with weight sharing. Maximum value 4
        num_branches (int): Number of branches after weight-shared blocks
        pytorch_weights_paths (Dict[str,int]): Strings corresponding to official imagenet resnet weights from pytorch
        resnetinput (ResnetInput): The input block consisting of conv layer, relu, and pooling.
        sharedblock (Union[nn.Sequential,nn.Identity]): The shared layers, consisting of at most 4 Resnet blocks.
        branches (nn.ModuleList): List of branching layers, with at most 4 Resnet blocks each.
    """

    block: nn.Module
    attention: str
    input_attention: bool
    ia_attention: bool
    part_attention: bool
    secondary_attention: int

    shared_block_count: int
    num_branches: int
    pytorch_weights_paths: Dict[str, int]

    resnetinput: ResnetInput
    sharedblock: nn.Sequential
    branches: nn.ModuleList

    def __init__(
        self,
        block: nn.Module = ResnetBottleneck,
        layers: List[int] = [3, 4, 6, 3],
        last_stride: int = 2,
        zero_init_residual: bool = False,
        top_only: bool = True,
        num_classes: bool = 1000,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: List[int] = None,
        norm_layer: nn.Module = None,
        attention: str = None,
        input_attention: bool = None,
        secondary_attention: int = None,
        ia_attention: bool = None,
        part_attention: bool = None,
        num_branches: int = 2,
        shared_block: int = 0,
        **kwargs
    ):
        """Initializes the multibranchresnet model and sets up internal modules.

        Args:
            block (nn.Module, optional): The block (BasicBlock or Bottleneck) used for the internal Resnet architecture. Defaults to ResnetBottleneck.
            layers (List[int], optional): Number of layers in each block. Defaults to [3, 4, 6, 3].
            last_stride (int, optional): The stride for the last block. Defaults to 2.
            zero_init_residual (bool, optional): Whether to initialize network with only zeros. Unused. Defaults to False.
            top_only (bool, optional): Whether to keep only the feature extractor block. Defaults to True.
            num_classes (bool, optional): Number of classes for the imagenet layer. Unused. Defaults to 1000.
            groups (int, optional): see nn.conv2D. Defaults to 1.
            width_per_group (int, optional): see nn.conv2D. Defaults to 64.
            replace_stride_with_dilation (List[int], optional): Whether to replace stride with dilation for each block. Defaults to None.
            norm_layer (nn.Module, optional): The default normalization layer. If None, uses batchnorm. Defaults to None.
            attention (str, optional): What attention to use, among `cbam`, `dbam`. Defaults to None.
            input_attention (bool, optional): Whether to use input attention at the first resnet block. Defaults to None.
            secondary_attention (int, optional): Whether to use secondary attention to apply `attention` to the specific resnet block only. Defaults to None.
            ia_attention (bool, optional): Whether to use input attention at the first conv layer. Exclusive with `input_attention`. Defaults to None.
            part_attention (bool, optional): Whether to use the local attention module. Defaults to None.
            num_branches (int, optional): Number of branches for this resnet. Defaults to 2.
            shared_block (int, optional): Number of weight-shared resnet blocks. Defaults to 0.

        Raises:
            ValueError: If attention blocks are not self-consistent. Specifically, the following rules:
                - cannot have both `ia_attention` and `input_attention`.
                - cannot have `part_attention` with `attention`, unless `secondary_attention`!=1
            ValueError: If branching does not occur, i.e. `shared_block`>=4
            ValueError: If `replace_stride_with_dilation` is not a 3-tuple or None.
        """
        super().__init__()

        self.pytorch_weights_paths = self._model_weights()
        self.block = block
        self.inplanes = 64
        if norm_layer is None:
            self._norm_layer = nn.BatchNorm2d
        # elif norm_layer == "ln":
        #    self._norm_layer = nn.LayerNorm
        self.dilation = 1
        if replace_stride_with_dilation is None:
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be `None` or a 3-element tuple. Got {}".format(
                    replace_stride_with_dilation
                )
            )
        self.groups = groups
        self.base_width = width_per_group

        # Attention parameters
        self.attention = attention
        self.input_attention = input_attention
        self.ia_attention = ia_attention
        self.part_attention = part_attention
        self.secondary_attention = secondary_attention

        # Make sure ia and input_attention do not conflict
        if self.ia_attention is not None and self.input_attention is not None:
            raise ValueError("Cannot have both ia_attention and input_attention.")
        if self.part_attention is not None and (
            self.attention is not None and self.secondary_attention is None
        ):
            raise ValueError("Cannot have part-attention with CBAM everywhere")
        if self.part_attention is not None and (
            self.attention is not None and self.secondary_attention == 1
        ):
            raise ValueError("Cannot have part-attention with CBAM-Early")

        # Here, set up where the branching begins in the resnet backbone
        self.shared_block_count = shared_block
        if self.shared_block_count > 4:
            raise ValueError(
                "`shared_block_count` value is %i. Cannot be greater than 4"
                % self.shared_block_count
            )
        if self.shared_block_count == 4:
            raise ValueError(
                "`shared_block_count` value is %i. This is a non-branching model."
                % self.shared_block_count
            )

        # Set up the per-layer parameters for the primary ResNet blocks
        layer_strides = [1, 2, 2, last_stride]
        layer_part_attention = [self.part_attention, False, False, False]
        layer_input_attention = [self.input_attention, False, False, False]
        layer_dilate = [False] + replace_stride_with_dilation
        layer_outplanes = [64, 128, 256, 512]
        # Fix secondary attention
        if secondary_attention is None:
            layer_att = [self.attention] * 4
        else:
            layer_att = [None] * 4
            layer_att[secondary_attention] = self.attention
        # Zip layer arguments
        layer_arguments = list(
            zip(
                layers,
                layer_outplanes,
                layer_strides,
                layer_part_attention,
                layer_input_attention,
                layer_dilate,
                layer_att,
            )
        )

        # First, given the shared_block_count, generate the shared layers list. We will nn.Sequential them later
        sharedlayers = []
        for layer_zip in layer_arguments[: self.shared_block_count]:
            sharedlayers.append(
                self._make_layer(
                    self.block,
                    layer_zip[1],
                    layer_zip[0],
                    attention=layer_zip[6],
                    input_attention=layer_zip[4],
                    part_attention=layer_zip[3],
                    dilate=layer_zip[5],
                    stride=layer_zip[2],
                )
            )
        self.shared_inplanes = self.inplanes
        self.shared_dilation = self.dilation
        # Then, given the branches, put the remaining resnet blocks in their branches
        # During prediction, we will just get branch features so order does not matter yet. it will matter in MultiBranchResnet
        # So, self.branches will be a nn.moduleList, with a bunch of nn.Sequentials
        self.num_branches = num_branches
        branches = [None] * self.num_branches
        for bidx in range(self.num_branches):
            branches[bidx] = []
            for layer_zip in layer_arguments[self.shared_block_count :]:
                branches[bidx].append(
                    self._make_layer(
                        self.block,
                        layer_zip[1],
                        layer_zip[0],
                        attention=layer_zip[6],
                        input_attention=layer_zip[4],
                        part_attention=layer_zip[3],
                        dilate=layer_zip[5],
                        stride=layer_zip[2],
                    )
                )
            branches[bidx] = nn.Sequential(*branches[bidx])
            self.inplanes = self.shared_inplanes
            self.dilation = self.shared_dilation

        self.resnetinput = ResnetInput(ia_attention=ia_attention)
        if len(sharedlayers) > 0:
            self.sharedblock = nn.Sequential(*sharedlayers)
        else:
            self.sharedblock = nn.Identity()
        self.branches = nn.ModuleList(branches)

    def _make_layer(
        self,
        block: nn.Module,
        planes: int,
        blocks: int,
        stride: int = 1,
        dilate: bool = False,
        attention: str = None,
        input_attention: bool = False,
        ia_attention: bool = False,
        part_attention: bool = False,
    ) -> nn.Sequential:
        """Creates a resnet block

        Args:
            block (nn.Module): The block (BasicBlock or Bottleneck) used for the internal Resnet architecture. Defaults to ResnetBottleneck.
            planes (int): Number of input depth
            blocks (int): Number of blocks in this ResnetBlock
            stride (int, optional): Stride for the conv layers. Defaults to 1.
            dilate (bool, optional): Dilation for the conv layers. Defaults to False.
            attention (str, optional): Which of `cbam`, `dbam` attention to use. Defaults to None.
            input_attention (bool, optional): Whether to use `input_attention`. Defaults to False.
            ia_attention (bool, optional): Whether to use `ia_attention`. Unused. Defaults to False.
            part_attention (bool, optional): Whether to use local attention. Defaults to False.

        Returns:
            nn.Sequential: The layers comprising this Resnet Block as an nn.Sequential
        """
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(
                    self.inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                self._norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                groups=self.groups,
                base_width=self.base_width,
                dilation=previous_dilation,
                norm_layer=self._norm_layer,
                attention=attention,
                input_attention=input_attention,
                part_attention=part_attention,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=self._norm_layer,
                    attention=attention,
                )
            )
        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.resnetinput(x)
        x = self.sharedblock(x)
        return [self.branches[idx](x) for idx in range(self.num_branches)]

    def load_param(self, weights_path: str):
        """Loads parameters from saved weights file

        Args:
            weights_path (str): Path to the weights file
        """
        if os.path.basename(weights_path) in self.pytorch_weights_paths:
            self.load_params_from_pytorch(weights_path)
        else:
            self.load_params_from_weights(weights_path)

    def load_params_from_pytorch(self, weights_path: str):
        """Loads default pytorch weights file into the multibranch resnet

        Args:
            weights_path (str): Path to the weights file
        """
        param_dict = torch.load(weights_path)
        # Three stages: load resnetinput params, load shared block params, and then load branch params...
        inputparams = [
            "conv1.weight",
            "bn1.running_mean",
            "bn1.running_var",
            "bn1.weight",
            "bn1.bias",
        ]
        # Load the input params
        for param in inputparams:
            self.state_dict()["resnetinput." + param].copy_(param_dict[param])

        # load shared block params,
        for layer_idx in range(self.shared_block_count):
            # layer_idx is the layer that is in shared-block. BUT, in the pytorch params, it exists as layer1, corresponding to layer_idx0
            # So if shared_block_count is 3, then we need to copy layer 1-3 into layer_idx0-2
            full_layer_list = [
                item for item in param_dict if ("layer" + str(layer_idx + 1) in item)
            ]  # get all weights inside layer[x]

            # The layers exist as an nn.sequential
            for layer_name in full_layer_list:
                # First, get the raw layer info, then append sharedblock to it...
                local_param_name = self._build_local_layer_param_from_pytorch_name(
                    layer_name, layer_idx
                )
                self.state_dict()[local_param_name].copy_(param_dict[layer_name])

        # Now, we need to load branch params...
        for layer_idx in range(self.shared_block_count, 4):
            # layer_idx is the layer that is in shared-block. BUT, in the pytorch params, it exists as layer1, corresponding to layer_idx0
            # So if shared_block_count is 3, then we need to copy layer 1-3 into layer_idx0-2
            full_layer_list = [
                item for item in param_dict if ("layer" + str(layer_idx + 1) in item)
            ]  # get all weights inside layer[x]

            # The layers exist as an nn.sequential
            for branch_idx in range(self.num_branches):
                for layer_name in full_layer_list:
                    # First, get the raw layer info, then append sharedblock to it...
                    local_param_name = self._build_branch_layer_param_from_pytorch_name(
                        layer_name, layer_idx, branch_idx, self.shared_block_count
                    )
                    self.state_dict()[local_param_name].copy_(param_dict[layer_name])

    def _build_local_layer_param_from_pytorch_name(
        self, paramname: str, layeridx: int
    ) -> str:
        """Converts a parameter name from an unbranched pytorch model to the corresponding `multibranchresnet` version inside the shared block.

        Args:
            paramname (str): The original parameter name
            layeridx (int): The layer this parameter name originated from

        Returns:
            str: Converted parameter name
        """
        param_list = paramname.split(".")
        new_param_list = ["sharedblock", str(layeridx)] + param_list[1:]
        return ".".join(new_param_list)

    def _build_branch_layer_param_from_pytorch_name(
        self, paramname: str, layeridx: int, branch_idx: int, layer_reset: int
    ) -> str:
        """Converts a parameter name from an unbranched pytorch model to the corresponding `multibranchresnet` version inside the branches.

        Args:
            paramname (str): The original parameter name
            layeridx (int): The layer this parameter name originated from
            branch_idx (int): The target branch for this parameter
            layer_reset (int): The layer reset number, since `multibranchresnet` restarts layer numbering at the branch junction

        Returns:
            str: Converted parameter name
        """
        param_list = paramname.split(".")
        new_param_list = [
            "branches",
            str(branch_idx),
            str(layeridx - layer_reset),
        ] + param_list[1:]
        return ".".join(new_param_list)

    def load_params_from_weights(self, weights_path: str):
        """Loads parameters from a saved `multibranchresnet`

        Args:
            weights_path (str): Path to the weights file
        """
        param_dict = torch.load(weights_path)
        for i in param_dict:
            if "fc" in i and self.top_only:
                continue
            self.state_dict()[i].copy_(param_dict[i])

    def _model_weights(self) -> Dict[str, int]:
        """Constructs a dictionary to quickly retrieve pytorch weight paths

        Returns:
            Dict[str,int]: the dictinary storing paths
        """
        mw = [
            "resnet18-5c106cde.pth",
            "resnet34-333f7ec4.pth",
            "resnet50-19c8e357.pth",
            "resnet101-5d3b4d8f.pth",
            "resnet152-b121ed2d.pth",
            "resnext50_32x4d-7cdf4587.pth",
            "resnext101_32x8d-8ba56ff5.pth",
            "wide_resnet50_2-95faca4d.pth",
            "wide_resnet50_2-95faca4d.pth",
            "resnet18-5c106cde_cbam.pth",
            "resnet34-333f7ec4_cbam.pth",
            "resnet50-19c8e357_cbam.pth",
            "resnet101-5d3b4d8f_cbam.pth",
            "resnet152-b121ed2d_cbam.pth",
        ]
        return {item: 1 for item in mw}


def _multibranchresnet(
    arch, block, layers, pretrained, progress, **kwargs
) -> multibranchresnet:
    """Builds a `multibranchresnet`

    Args:
        arch (str): Architecture base. Unused.
        block (nn.Module): The class of the ResnetBlock (BasicBlock or Bottleneck)
        layers (List[int]): The layers for each ResnetBlock
        pretrained (bool): Unused
        progress (bool): Unused

    Returns:
        multibranchresnet: The `multibranchresnet` model
    """
    model = multibranchresnet(block, layers, **kwargs)
    return model


def resnet18(pretrained=False, progress=True, **kwargs):
    r"""ResNet-18 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _multibranchresnet(
        "resnet18", ResnetBasicBlock, [2, 2, 2, 2], pretrained, progress, **kwargs
    )


def resnet34(pretrained=False, progress=True, **kwargs):
    r"""ResNet-34 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _multibranchresnet(
        "resnet34", ResnetBasicBlock, [3, 4, 6, 3], pretrained, progress, **kwargs
    )


def resnet50(pretrained=False, progress=True, **kwargs):
    r"""ResNet-50 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _multibranchresnet(
        "resnet50", ResnetBottleneck, [3, 4, 6, 3], pretrained, progress, **kwargs
    )


def resnet101(pretrained=False, progress=True, **kwargs):
    r"""ResNet-101 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _multibranchresnet(
        "resnet101", ResnetBottleneck, [3, 4, 23, 3], pretrained, progress, **kwargs
    )


def resnet152(pretrained=False, progress=True, **kwargs):
    r"""ResNet-152 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _multibranchresnet(
        "resnet152", ResnetBottleneck, [3, 8, 36, 3], pretrained, progress, **kwargs
    )


def resnext50_32x4d(pretrained=False, progress=True, **kwargs):
    r"""ResNeXt-50 32x4d model from
    `"Aggregated Residual Transformation for Deep Neural Networks" <https://arxiv.org/pdf/1611.05431.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 4
    return _multibranchresnet(
        "resnext50_32x4d",
        ResnetBottleneck,
        [3, 4, 6, 3],
        pretrained,
        progress,
        **kwargs
    )


def resnext101_32x8d(pretrained=False, progress=True, **kwargs):
    r"""ResNeXt-101 32x8d model from
    `"Aggregated Residual Transformation for Deep Neural Networks" <https://arxiv.org/pdf/1611.05431.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 8
    return _multibranchresnet(
        "resnext101_32x8d",
        ResnetBottleneck,
        [3, 4, 23, 3],
        pretrained,
        progress,
        **kwargs
    )


def wide_resnet50_2(pretrained=False, progress=True, **kwargs):
    r"""Wide ResNet-50-2 model from
    `"Wide Residual Networks" <https://arxiv.org/pdf/1605.07146.pdf>`_
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["width_per_group"] = 64 * 2
    return _multibranchresnet(
        "wide_resnet50_2",
        ResnetBottleneck,
        [3, 4, 6, 3],
        pretrained,
        progress,
        **kwargs
    )


def wide_resnet101_2(pretrained=False, progress=True, **kwargs):
    r"""Wide ResNet-101-2 model from
    `"Wide Residual Networks" <https://arxiv.org/pdf/1605.07146.pdf>`_
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["width_per_group"] = 64 * 2
    return _multibranchresnet(
        "wide_resnet101_2",
        ResnetBottleneck,
        [3, 4, 23, 3],
        pretrained,
        progress,
        **kwargs
    )



#!/usr/bin/env python3
"""Generators yields batches to models during training and testing. 
Requirs a crawler to provide a list of paths with labels.
Generator then loads a random subset, performs transformations, 
and yields these to model when requested.

List of supported generators:

    - SequencedGenerator: Yields a triplet batch for re-id
    - TripletGenerator: Yields a triplet batch for re-id
    - ClassedGenerator:
    - Cars196Generator:
    - ClassificationGenerator:
    - CoLabelIntegratedDatasetGenerator: 
    - CoLabelDeployGenerator:
    - KnowledgeIntegratedGenerator: 
"""

from typing import Any, List, Tuple, Union
import torchvision.transforms as T
from torch.utils.data import Dataset as TorchDataset

import ednaml.utils
from ednaml.utils.LabelMetadata import LabelMetadata


class ImageGenerator:
    """Base class for image dataset generators
    """

    num_entities: LabelMetadata

    def __init__(
        self,
        gpus: int,
        i_shape: Union[List[int], Tuple[int, int]],
        channels: int,
        normalization_mean: float,
        normalization_std: float,
        normalization_scale: float,
        **kwargs
    ):
        """Initializes the Generator and builds the data transformer

        Args:
            gpus (_type_): _description_
            i_shape (_type_): _description_
            normalization_mean (_type_): _description_
            normalization_std (_type_): _description_
            normalization_scale (_type_): _description_
        """
        self.gpus = gpus
        self.transformer = T.Compose(
            self.build_transforms(
                i_shape,
                channels,
                normalization_mean,
                normalization_std,
                normalization_scale,
                **kwargs
            )
        )

    def build_transforms(
        self,
        i_shape: Union[List[int], Tuple[int, int]],
        channels: int,
        normalization_mean: float,
        normalization_std: float,
        normalization_scale: float,
        **kwargs
    ) -> List[object]:
        """Builds the transforms for the images in dataset. This can be replaced for custom set of transforms

        Args:
            i_shape (Union[List[int,int],Tuple[int,int]]): _description_
            normalization_mean (float): _description_
            normalization_std (float): _description_
            normalization_scale (float): _description_

        Returns:
            _type_: _description_
        """
        normalization_mean, normalization_std = ednaml.utils.extend_mean_arguments(
            [normalization_mean, normalization_std], channels
        )
        transformer_primitive = []
        transformer_primitive.append(T.Resize(size=i_shape))
        if kwargs.get("h_flip") > 0:
            transformer_primitive.append(T.RandomHorizontalFlip(p=kwargs.get("h_flip")))
        if kwargs.get("t_crop"):
            transformer_primitive.append(T.RandomCrop(size=i_shape))
        transformer_primitive.append(T.ToTensor())
        transformer_primitive.append(
            T.Normalize(mean=normalization_mean, std=normalization_std)
        )
        if kwargs.get("rea"):
            transformer_primitive.append(
                T.RandomErasing(
                    p=0.5, scale=(0.02, 0.4), value=kwargs.get("rea_value", 0)
                )
            )
        return transformer_primitive

    # NOTE removed instance parameter from here.,, is it needed???
    def setup(self, datacrawler, mode, batch_size, workers, **kwargs):
        """This should generate a TorchDataset and associated DataLoader to yield batches.
        The actual steps are as follows:

        Raises:
            NotImplementedError: _description_
        """

        self.workers = workers * self.gpus

        self.dataset = self.buildDataset(datacrawler, mode, self.transformer, **kwargs)
        self.dataloader = self.buildDataLoader(
            self.dataset, mode, batch_size=batch_size, **kwargs
        )
        self.num_entities = self.getNumEntities(datacrawler, mode, **kwargs)

    def buildDataset(
        self, datacrawler, mode: str, transform: List[object], **kwargs
    ) -> TorchDataset:
        """Given the daatacrawler with all the data, and the mode (could be 
        any user-defined mode such as 'train', 'test', 'zsl', 'gzsl', etc), as 
        as well the transform, return a TorchDataset

        Args:
            datacrawler (_type_): _description_
            mode (_type_): _description_
            transform (_type_): _description_

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError()

    def buildDataLoader(self, dataset, mode, batch_size, **kwargs):
        """Given a Torch Dataset, build a TorchDataloader to return a tensor, possibly with a collate function

        Args:
            dataset (_type_): _description_
            batch_size (_type_): _description_

        Raises:
            NotImplementedError: _description_
        """

        raise NotImplementedError()

    def getNumEntities(self, datacrawler, mode, **kwargs) -> LabelMetadata:
        """Return the number of classes in the overall label scheme
        Raises:
            NotImplementedError: _description_
        """

        raise NotImplementedError()


from .ClassificationGenerator import ClassificationGenerator
from .CoLabelIntegratedDatasetGenerator import CoLabelIntegratedDatasetGenerator
from .CoLabelDeployGenerator import CoLabelDeployGenerator
from .MultiClassificationGenerator import MultiClassificationGenerator

KnowledgeIntegratedGenerator = CoLabelIntegratedDatasetGenerator



# Proposed Splits for SUN from
# Zero-shot Learning - The Good, the Bad and the Ugly. Y. Xian, B. Schiele, Z. Akata. IEEE CVPR 2017.

trainval = r"""bakery_shop
bar
barn
boathouse
bookstore
bridge
butchers_shop
candy_store
carrousel
castle
cliff
closet
coast
dam
dock
fastfood_restaurant
fly_bridge
fountain
home_theater
house
lift_bridge
lighthouse
manufactured_home
mobile_home
nursery
palace
patio
pier
restaurant
restaurant_patio
roof
rope_bridge
sandbar
sawmill
sea_cliff
seawall
roundabout
toyshop
trestle_bridge
underwater_coral_reef
underwater_wreck
valley
viaduct
volcano
water_tower
vegetable_garden
chemical_plant
residential_neighborhood
tent_indoor
utility_room
swimming_hole
darkroom
shipping_room
shed
gangplank
forest_broadleaf
ferryboat_outdoor
campus
railroad_track
auto_racing_paddock
zoo
garage_outdoor
strip_mine
fish_farm
airfield
glacier
dolmen
bistro_outdoor
music_store
diner_outdoor
general_store_outdoor
bullpen
barndoor
discotheque
poolroom_home
warehouse_indoor
art_studio
thriftshop
campsite
golf_course
jail_indoor
train_railway
carport_outdoor
atrium_public
museum_outdoor
woodland
fire_escape
fjord
brewery_outdoor
sun_deck
balcony_interior
videostore
underwater_kelp_forest
ski_lodge
cybercafe
tent_outdoor
flea_market_indoor
day_care_center
village
mesa
mastaba
subway_interior
jury_box
earth_fissure
herb_garden
chalet
beer_garden
conference_center
wharf
food_court
drill_rig
throne_room
hunting_lodge_outdoor
control_room
dinette_home
military_hut
berth
kitchen
water_treatment_plant_outdoor
baggage_claim
fence
clothing_store
library_outdoor
mission
revolving_door
sand_trap
amphitheater
hill
aquatic_theater
cathedral_indoor
nursing_home
tea_garden
editing_room
dinette_vehicle
gazebo_exterior
limousine_interior
forest_needleleaf
basketball_court_outdoor
martial_arts_gym
brickyard_outdoor
donjon
trench
particle_accelerator
bus_shelter
clean_room
banquet_hall
weighbridge
caravansary
rainforest
recycling_plant_outdoor
newsroom
assembly_line
flood
elevator_door
atrium_home
oilrig
call_center
estuary
pulpit
office_cubicles
checkout_counter
van_interior
ski_slope
apse_outdoor
morgue
ice_skating_rink_indoor
chicken_farm_indoor
raceway
mini_golf_course_outdoor
shower
stone_circle
subway_station_corridor
market_outdoor
velodrome_outdoor
pharmacy
power_plant_indoor
canteen
dining_hall
desert_vegetation
overpass
drainage_ditch
print_shop
circus_tent_outdoor
steel_mill_indoor
bank_indoor
butte
kindergarden_classroom
ramp
waterfall_cascade
military_hospital
rolling_mill
conference_hall
baseball_field
boat_deck
subway_station_platform
control_tower_indoor
schoolyard
bus_station_outdoor
snowfield
anechoic_chamber
watchtower
formal_garden
sewing_room
catwalk
burial_chamber
alcove
carport_freestanding
airport_ticket_counter
embassy
office
stable
liquor_store_indoor
arena_performance
raft
kiosk_outdoor
resort
cabin_outdoor
pizzeria
dugout
courthouse
movie_theater_indoor
physics_laboratory
cardroom
tennis_court_outdoor
stage_indoor
badlands
lake_natural
waiting_room
pitchers_mound
pantry
lido_deck_outdoor
shipyard_outdoor
escalator_outdoor
crawl_space
wine_cellar_barrel_storage
cathedral_outdoor
temple_east_asia
theater_indoor_procenium
cottage
dry_dock
hallway
outcropping
parking_garage_indoor
ice_skating_rink_outdoor
phone_booth
ghost_town
factory_outdoor
bedchamber
botanical_garden
chicken_coop_outdoor
corn_field
pavilion
firing_range_outdoor
bamboo_forest
industrial_area
pig_farm
gatehouse
restroom_outdoor
construction_site
gasworks
dacha
nuclear_power_plant_indoor
field_road
lab_classroom
attic
landfill
driving_range_outdoor
fire_station
corridor
restaurant_kitchen
underwater_ocean_deep
witness_stand
aquarium
manhole
courtroom
poolroom_establishment
hotel_outdoor
bedroom
kasbah
basilica
skatepark
joss_house
nuclear_power_plant_outdoor
jewelry_shop
tunnel_rail_outdoor
mansion
bakery_kitchen
plantation_house
library_indoor
basketball_court_indoor
oast_house
cloister_outdoor
underwater_pool
tower
courtyard
portico
freeway
gift_shop
childs_room
cloister_indoor
quay
gulch
bell_foundry
moat_water
conference_room
driveway
grotto
pagoda
television_studio
escalator_indoor
pilothouse_outdoor
theater_indoor_round
ice_floe
levee
art_school
booth_indoor
control_tower_outdoor
office_building
hut
swamp
liquor_store_outdoor
badminton_court_outdoor
auto_showroom
oasis
plaza
inn_outdoor
youth_hostel
canal_urban
veranda
synagogue_indoor
ski_jump
desert_sand
park
waterfall_cataract
guardhouse
repair_shop
pond
factory_indoor
pump_room
hedge_maze
restroom_indoor
beer_hall
beach_house
riding_arena
podium_indoor
balcony_exterior
field_wild
soccer_field
organ_loft_exterior
desert_road
hot_spring
airport_airport
hospital_room
topiary_garden
hot_tub_indoor
podium_outdoor
badminton_court_indoor
science_museum
farm
amusement_arcade
cafeteria
downtown
baptistry_indoor
parlor
heliport
garbage_dump
theater_outdoor
train_station_outdoor
gymnasium_indoor
packaging_plant
ball_pit
fortress
bowling_alley
sky
heath
airport_terminal
lagoon
cargo_deck_airplane
mine
power_plant_outdoor
mausoleum
aqueduct
reading_room
quonset_hut_outdoor
computer_room
kennel_outdoor
catacomb
waterfall_plunge
junkyard
tree_house
labyrinth_indoor
highway
bleachers_outdoor
game_room
quadrangle
water_mill
lobby
confessional
salt_plain
shoe_shop
mountain_path
airport_entrance
cabana
apartment_building_outdoor
access_road
batting_cage_indoor
mosque_outdoor
airplane_cabin
putting_green
kennel_indoor
archaelogical_excavation
natural_history_museum
recreation_room
submarine_interior
window_seat
teashop
shopping_mall_indoor
railway_yard
wine_cellar_bottle_storage
hacienda
living_room
chaparral
movie_theater_outdoor
pasture
hospital
casino_indoor
elevator_shaft
water_treatment_plant_indoor
shopfront
cavern_indoor
root_cellar
priory
armory
stadium_outdoor
stadium_football
greenhouse_outdoor
bicycle_racks
planetarium_outdoor
gun_deck_indoor
wave
engine_room
boardwalk
bullring
dentists_office
classroom
terrace_farm
drugstore
athletic_field_outdoor
orchard
beach
junk_pile
auto_factory
pub_outdoor
toll_plaza
observatory_indoor
building_facade
swimming_pool_outdoor
locker_room
lecture_room
inn_indoor
car_interior_backseat
stadium_baseball
jail_outdoor
doorway_outdoor
lean-to
veterinarians_office
flea_market_outdoor
baptistry_outdoor
strip_mall
boxing_ring
supermarket
biology_laboratory
ruin
sacristy
beauty_salon
mountain_road
ocean
chicken_coop_indoor
creek
cottage_garden
harbor
hayfield
rice_paddy
pilothouse_indoor
jacuzzi_outdoor
marsh
dining_car
velodrome_indoor
ticket_window_outdoor
home_office
parade_ground
mineshaft
gas_station
sauna
greenhouse_indoor
departure_lounge
pedestrian_overpass_outdoor
winery
pet_shop
zen_garden
arch
forest_road
football_field
batters_box
chicken_farm_outdoor
runway
elevator_lobby
track_indoor
windmill
lake_artificial
apse_indoor
lock_chamber
ranch_house
medina
underwater_ocean_shallow
crevasse
outhouse_outdoor
garage_indoor
skyscraper
amusement_park
wind_farm
recycling_plant_indoor
promenade
moor
chapel
street
kiosk_indoor
slum
bayou
temple_western
porch
river
loading_dock
general_store_indoor
tollbooth
museum_indoor
hedgerow
voting_booth
islet
waterfall_block
machine_shop
foundry_indoor
igloo
irrigation_ditch
staircase
hunting_lodge_indoor
coffee_shop
dairy_indoor
auto_mechanics_indoor
stage_outdoor
dorm_room
circus_tent_indoor
tree_farm
market_indoor
fairway
moat_dry
bus_interior
newsstand_outdoor
optician
tearoom
convenience_store_indoor
road_cut
underwater_ice
jail_cell
crosswalk
iceberg
naval_base
bow_window_outdoor
reception
tennis_court_indoor
track_outdoor
lookout_station_outdoor
hot_tub_outdoor
town_house
picnic_area
arrival_gate_outdoor
arena_hockey
train_depot
covered_bridge_exterior
mountain_snowy
backstage
signal_box
flight_of_stairs_natural
art_gallery
cockpit
great_hall
stilt_house_water
operating_room
diner_indoor
rock_arch
ventilation_shaft
kitchenette
furnace_room
church_outdoor
bistro_indoor
trailer_park
spillway
oil_refinery_outdoor
waterfall_fan
labyrinth_outdoor
playroom
bindery
building_complex
flight_of_stairs_urban
ranch
laundromat
wheat_field
workroom
rubble
mountain
hangar_outdoor
delicatessen
elevator_freight_elevator
basement
schoolhouse
city
fort
abbey
steel_mill_outdoor
wet_bar
squash_court
geodesic_dome_outdoor
parking_garage_outdoor
electrical_substation
sushi_bar
bank_outdoor
planetarium_indoor
florist_shop_indoor
foundry_outdoor
stadium_soccer
funeral_chapel
canyon
music_studio
gorge
hotel_breakfast_area
train_station_station
industrial_park
cheese_factory
dining_room
swimming_pool_indoor
server_room
ossuary
synagogue_outdoor
forest_path
bathroom"""

query = r"""airlock
temple_south_asia
monastery_outdoor
piano_store
watering_hole
geodesic_dome_indoor
chemistry_lab
vineyard
brewery_indoor
pub_indoor
vestry
bog
wrestling_ring_indoor
trading_floor
volleyball_court_outdoor
racecourse
cemetery
betting_shop
casino_outdoor
theater_indoor_seats
dirt_track
motel
ballroom
sandbox
mosque_indoor
batting_cage_outdoor
bow_window_indoor
car_interior_frontseat
alley
nightclub
arena_basketball
hotel_room
jacuzzi_indoor
tunnel_road_outdoor
bank_vault
excavation
lawn
galley
bazaar_outdoor
canal_natural
train_station_platform
cubicle_office
fishpond
ice_shelf
bazaar_indoor
field_cultivated
exhibition_hall
observatory_outdoor
rectory
tundra
firing_range_indoor
doorway_indoor
ticket_booth
bus_depot_outdoor
auditorium
ski_resort
ziggurat
elevator_interior
artists_loft
archive
playground
hoodoo
parking_lot
yard
corral
japanese_garden
promenade_deck
landing_deck
workshop
hangar_indoor
church_indoor
savanna"""

trainval = trainval.split("\n")
query = query.split("\n")



from ednaml.config.ConfigDefaults import ConfigDefaults
from ednaml.config.ExecutionDatareaderConfig import ExecutionDatareaderConfig


class ExecutionConfig:
    OPTIMIZER_BUILDER: str
    MODEL_SERVING: bool
    EPOCHS: int
    SKIPEVAL: bool
    TEST_FREQUENCY: int
    TRAINER: str
    DATAREADER: ExecutionDatareaderConfig

    def __init__(self, execution_dict, defaults: ConfigDefaults):
        self.OPTIMIZER_BUILDER = execution_dict.get(
            "OPTIMIZER_BUILDER", defaults.OPTIMIZER_BUILDER
        )
        self.MODEL_SERVING = execution_dict.get("MODEL_SERVING", defaults.MODEL_SERVING)
        self.EPOCHS = execution_dict.get("EPOCHS", defaults.EPOCHS)
        self.SKIPEVAL = execution_dict.get("SKIPEVAL", defaults.SKIPEVAL)
        self.TEST_FREQUENCY = execution_dict.get(
            "TEST_FREQUENCY", defaults.TEST_FREQUENCY
        )
        self.TRAINER = execution_dict.get("TRAINER", defaults.TRAINER)

        self.DATAREADER = ExecutionDatareaderConfig(
            execution_dict.get("DATAREADER", {})
        )



import torch
from torch import nn
import torch.nn.init
import torch.nn.functional as F
from torch.nn.modules.batchnorm import _BatchNorm


class LambdaLayer(nn.Module):
    """ Torch lambda layer to act as an empty layer. It does not do anything """

    def __init__(self, lambd):
        super(LambdaLayer, self).__init__()
        self._lambda = lambd

    def forward(self, x):
        return self._lambda(x)


# https://github.com/amdegroot/ssd.pytorch/blob/master/layers/modules/l2norm.py
class L2Norm(nn.Module):
    def __init__(self, n_channels, scale):
        super(L2Norm, self).__init__()
        self.n_channels = n_channels
        self.gamma = scale or None
        self.eps = 1e-10
        self.weight = nn.Parameter(torch.Tensor(self.n_channels))
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.constant_(self.weight, self.gamma)

    def forward(self, x):
        norm = x.pow(2).sum(dim=1, keepdim=True).sqrt() + self.eps
        x = torch.div(x, norm)
        out = self.weight.expand_as(x) * x
        return out


from torch.nn.modules import instancenorm

# https://github.com/pytorch/pytorch/pull/9924/commits/816d048b91a455a5f57da6f1ab304e70c38a39bb
class FixedInstanceNorm1d(instancenorm._InstanceNorm):
    """Applies Instance Normalization over a 2D or 3D input (a mini-batch of 1D
    inputs with optional additional channel dimension) as described in the paper
    `Instance Normalization: The Missing Ingredient for Fast Stylization`_ .

    TODO Fix this formatting to match sphinx documentation

    .. math::
        y = \frac{x - \mathrm{E}[x]}{ \sqrt{\mathrm{Var}[x] + \epsilon}} * \gamma + \beta
    The mean and standard-deviation are calculated per-dimension separately
    for each object in a mini-batch. :math:`\gamma` and :math:`\beta` are learnable parameter vectors
    of size `C` (where `C` is the input size) if :attr:`affine` is ``True``.
    By default, this layer uses instance statistics computed from input data in
    both training and evaluation modes.
    If :attr:`track_running_stats` is set to ``True``, during training this
    layer keeps running estimates of its computed mean and variance, which are
    then used for normalization during evaluation. The running estimates are
    kept with a default :attr:`momentum` of 0.1.
    .. note::
        This :attr:`momentum` argument is different from one used in optimizer
        classes and the conventional notion of momentum. Mathematically, the
        update rule for running statistics here is
        :math:`\hat{x}_\text{new} = (1 - \text{momentum}) \times \hat{x} + \text{momemtum} \times x_t`,
        where :math:`\hat{x}` is the estimated statistic and :math:`x_t` is the
        new observed value.
    Args:
        num_features: :math:`C` from an expected input of size
            :math:`(N, C, L)` or :math:`L` from input of size :math:`(N, L)`
        eps: a value added to the denominator for numerical stability. Default: 1e-5
        momentum: the value used for the running_mean and running_var computation. Default: 0.1
        affine: a boolean value that when set to ``True``, this module has
            learnable affine parameters, initialized the same way as done for batch normalization.
            Default: ``False``.
        track_running_stats: a boolean value that when set to ``True``, this
            module tracks the running mean and variance, and when set to ``False``,
            this module does not track such statistics and always uses batch
            statistics in both training and eval modes. Default: ``False``
    Shape:
        - Input: :math:`(N, C, L)`
        - Output: :math:`(N, C, L)` (same shape as input)
    Examples::
        >>> # Without Learnable Parameters
        >>> m = nn.InstanceNorm1d(100)
        >>> # With Learnable Parameters
        >>> m = nn.InstanceNorm1d(100, affine=True)
        >>> input = torch.randn(20, 100, 40)
        >>> output = m(input)
    .. _`Instance Normalization: The Missing Ingredient for Fast Stylization`:
        https://arxiv.org/abs/1607.08022
    """

    def _check_input_dim(self, input):
        if input.dim() != 2 and input.dim() != 3:
            raise ValueError(
                "expected 2D or 3D input (got {}D input)".format(input.dim())
            )


def group_norm(
    input,
    group,
    running_mean,
    running_var,
    weight=None,
    bias=None,
    use_input_stats=True,
    momentum=0.1,
    eps=1e-5,
):
    r"""Applies Group Normalization for channels in the same group in each data sample in a
    batch.
    See :class:`~torch.nn.GroupNorm1d`, :class:`~torch.nn.GroupNorm2d`,
    :class:`~torch.nn.GroupNorm3d` for details.
    """
    if not use_input_stats and (running_mean is None or running_var is None):
        raise ValueError(
            "Expected running_mean and running_var to be not None when use_input_stats=False"
        )

    b, c = input.size(0), input.size(1)
    if weight is not None:
        weight = weight.repeat(b)
    if bias is not None:
        bias = bias.repeat(b)

    def _instance_norm(
        input,
        group,
        running_mean=None,
        running_var=None,
        weight=None,
        bias=None,
        use_input_stats=None,
        momentum=None,
        eps=None,
    ):
        # Repeat stored stats and affine transform params if necessary
        if running_mean is not None:
            running_mean_orig = running_mean
            running_mean = running_mean_orig.repeat(b)
        if running_var is not None:
            running_var_orig = running_var
            running_var = running_var_orig.repeat(b)

        # norm_shape = [1, b * c / group, group]
        # print(norm_shape)
        # Apply instance norm
        input_reshaped = input.contiguous().view(
            1, int(b * c / group), group, *input.size()[2:]
        )

        out = F.batch_norm(
            input_reshaped,
            running_mean,
            running_var,
            weight=weight,
            bias=bias,
            training=use_input_stats,
            momentum=momentum,
            eps=eps,
        )

        # Reshape back
        if running_mean is not None:
            running_mean_orig.copy_(
                running_mean.view(b, int(c / group)).mean(0, keepdim=False)
            )
        if running_var is not None:
            running_var_orig.copy_(
                running_var.view(b, int(c / group)).mean(0, keepdim=False)
            )

        return out.view(b, c, *input.size()[2:])

    return _instance_norm(
        input,
        group,
        running_mean=running_mean,
        running_var=running_var,
        weight=weight,
        bias=bias,
        use_input_stats=use_input_stats,
        momentum=momentum,
        eps=eps,
    )


class _GroupNorm(_BatchNorm):
    def __init__(
        self,
        num_features,
        num_groups=1,
        eps=1e-5,
        momentum=0.1,
        affine=False,
        track_running_stats=False,
    ):
        self.num_groups = num_groups
        self.track_running_stats = track_running_stats
        super(_GroupNorm, self).__init__(
            int(num_features / num_groups), eps, momentum, affine, track_running_stats
        )

    def _check_input_dim(self, input):
        return NotImplemented

    def forward(self, input):
        self._check_input_dim(input)

        return group_norm(
            input,
            self.num_groups,
            self.running_mean,
            self.running_var,
            self.weight,
            self.bias,
            self.training or not self.track_running_stats,
            self.momentum,
            self.eps,
        )


class GroupNorm2d(_GroupNorm):
    r"""Applies Group Normalization over a 4D input (a mini-batch of 2D inputs
    with additional channel dimension) as described in the paper
    https://arxiv.org/pdf/1803.08494.pdf
    `Group Normalization`_ .
    Args:
        num_features: :math:`C` from an expected input of size
            :math:`(N, C, H, W)`
        num_groups:
        eps: a value added to the denominator for numerical stability. Default: 1e-5
        momentum: the value used for the running_mean and running_var computation. Default: 0.1
        affine: a boolean value that when set to ``True``, this module has
            learnable affine parameters. Default: ``True``
        track_running_stats: a boolean value that when set to ``True``, this
            module tracks the running mean and variance, and when set to ``False``,
            this module does not track such statistics and always uses batch
            statistics in both training and eval modes. Default: ``False``
    Shape:
        - Input: :math:`(N, C, H, W)`
        - Output: :math:`(N, C, H, W)` (same shape as input)
    Examples:
        >>> # Without Learnable Parameters
        >>> m = GroupNorm2d(100, 4)
        >>> # With Learnable Parameters
        >>> m = GroupNorm2d(100, 4, affine=True)
        >>> input = torch.randn(20, 100, 35, 45)
        >>> output = m(input)
    """

    def _check_input_dim(self, input):
        if input.dim() != 4:
            raise ValueError("expected 4D input (got {}D input)".format(input.dim()))


class GroupNorm3d(_GroupNorm):
    """
        Assume the data format is (B, C, D, H, W)
    """

    def _check_input_dim(self, input):
        if input.dim() != 5:
            raise ValueError("expected 5D input (got {}D input)".format(input.dim()))



import torch
import torch.nn as nn
from ednaml.loss import Loss


class CenterLoss(Loss):
    """Center loss.
    
    Reference:
    Wen et al. A Discriminative Feature Learning Approach for Deep Face Recognition. ECCV 2016.
    
    Args:
        num_classes (int): number of classes.
        feat_dim (int): feature dimension.
    """

    def __init__(self, num_classes, feat_dim, use_gpu=True):
        super(CenterLoss, self).__init__()

        self.num_classes = num_classes
        self.feat_dim = feat_dim
        self.centers = nn.Parameter(torch.randn(self.num_classes, self.feat_dim).cuda())

    def forward(self, features, labels):
        """
        Args:
            features: feature matrix with shape (batch_size, feat_dim).
            labels: ground truth labels with shape (batch_size).
        """
        batch_size = features.size(0)
        distmat = (
            torch.pow(features, 2)
            .sum(dim=1, keepdim=True)
            .expand(batch_size, self.num_classes)
            + torch.pow(self.centers, 2)
            .sum(dim=1, keepdim=True)
            .expand(self.num_classes, batch_size)
            .t()
        )
        distmat.addmm_(1, -2, features, self.centers.t())

        classes = torch.arange(self.num_classes).long().cuda()
        labels = labels.unsqueeze(1).expand(batch_size, self.num_classes)
        mask = labels.eq(classes.expand(batch_size, self.num_classes))

        dist = distmat * mask.float()
        loss = dist.clamp(min=1e-12, max=1e12).sum() / batch_size

        return loss
